"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains periodograms functions for time series data.

Code written by: Luc IJspeert
"""

import numpy as np
import numba as nb
import astropy.timeseries as apy


# get the number of available cpu threads
n_proc = nb.get_num_threads()


@nb.njit(cache=True)
def fold_time_series_phase(time, p_orb, zero=None):
    """Fold the given time series over the orbital period to transform to phase space.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    p_orb: float
        The orbital period with which the time series is folded
    zero: float, None
        Reference zero point in time when the phase equals zero

    Returns
    -------
    numpy.ndarray[Any, dtype[float]]
        Phase array for all timestamps. Phases are between -0.5 and 0.5
    """
    mean_t = np.mean(time)
    if zero is None:
        zero = -mean_t
    phases = ((time - mean_t - zero) / p_orb + 0.5) % 1 - 0.5

    return phases


@nb.njit(cache=True)
def fold_time_series(time, p_orb, t_zero, t_ext_1=0, t_ext_2=0):
    """Fold the given time series over the orbital period

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    p_orb: float
        The orbital period with which the time series is folded
    t_zero: float, None
        Reference zero point in time (with respect to the time series mean time)
        when the phase equals zero
    t_ext_1: float
        Negative time interval to extend the folded time series to the left.
    t_ext_2: float
        Positive time interval to extend the folded time series to the right.

    Returns
    -------
    tuple
        A tuple containing the following elements:
        t_extended: numpy.ndarray[Any, dtype[float]]
            Folded time series array for all timestamps (and possible extensions).
        ext_left: numpy.ndarray[bool]
            Mask of points to extend time series to the left (for if t_ext_1!=0)
        ext_right: numpy.ndarray[bool]
            Mask of points to extend time series to the right (for if t_ext_2!=0)
    """
    # reference time is the mean of the time array
    mean_t = np.mean(time)
    t_folded = (time - mean_t - t_zero) % p_orb

    # extend to both sides
    ext_left = (t_folded > p_orb + t_ext_1)
    ext_right = (t_folded < t_ext_2)
    t_extended = np.concatenate((t_folded[ext_left] - p_orb, t_folded, t_folded[ext_right] + p_orb))

    return t_extended, ext_left, ext_right


@nb.njit(cache=True)
def phase_dispersion(phases, flux, n_bins):
    """Phase dispersion, as in PDM, without overlapping bins.

    Parameters
    ----------
    phases: numpy.ndarray[Any, dtype[float]]
        The phase-folded timestamps of the time series, between -0.5 and 0.5.
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    n_bins: int
        The number of bins over the orbital phase

    Returns
    -------
    float
        Phase dispersion, or summed variance over the bins divided by
        the variance of the flux

    Notes
    -----
    Intentionally does not make use of scipy to enable JIT-ting, which makes this considerably faster.
    """

    def var_no_avg(a):
        return np.sum(np.abs(a - np.mean(a)) ** 2)  # if mean instead of sum, this is variance

    edges = np.linspace(-0.5, 0.5, n_bins + 1)
    # binned, edges, indices = sp.stats.binned_statistic(phases, flux, statistic=statistic, bins=bins)
    binned_var = np.zeros(n_bins)
    for i, (b1, b2) in enumerate(zip(edges[:-1], edges[1:])):
        bin_mask = (phases >= b1) & (phases < b2)
        if np.any(bin_mask):
            binned_var[i] = var_no_avg(flux[bin_mask])
        else:
            binned_var[i] = 0

    total_var = np.sum(binned_var) / len(flux)
    overall_var = np.var(flux)

    return total_var / overall_var


@nb.njit(cache=True, parallel=True)
def phase_dispersion_minimisation(time, flux, f_n, local=False):
    """Determine the phase dispersion over a set of periods to find the minimum

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    local: bool
        If set True, only searches the given frequencies,
        else also fractions of the frequencies are searched

    Returns
    -------
    tuple
        A tuple containing the following elements:
        periods: numpy.ndarray[Any, dtype[float]]
            Periods at which the phase dispersion is calculated
        pd_all: numpy.ndarray[Any, dtype[float]]
            Phase dispersion at the given periods
    """
    # number of bins for dispersion calculation
    n_points = len(time)
    if n_points / 10 > 1000:
        n_bins = 1000
    else:
        n_bins = n_points // 10  # at least 10 data points per bin on average

    # determine where to look based on the frequencies, including fractions of the frequencies
    if local:
        periods = 1 / f_n
    else:
        periods = np.zeros(7 * len(f_n))
        for i, f in enumerate(f_n):
            periods[7 * i:7 * i + 7] = np.arange(1, 8) / f

    # stay below the maximum
    periods = periods[periods < np.ptp(time)]

    # and above the minimum
    periods = periods[periods > (2 * np.min(time[1:] - time[:-1]))]

    # compute the dispersion measures
    n_periods = len(periods)
    pd_all = np.zeros(n_periods)
    for i in nb.prange(n_periods):
        fold = fold_time_series_phase(time, periods[i], 0)
        pd_all[i] = phase_dispersion(fold, flux, n_bins)

    return periods, pd_all


def scargle_noise_spectrum(time, resid, window_width=1.0):
    """Calculate the Lomb-Scargle noise spectrum by a convolution with a flat window of a certain width.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    resid: numpy.ndarray[Any, dtype[float]]
        Residual measurement values of the time series
    window_width: float
        The width of the window used to compute the noise spectrum,
        in inverse unit of the time array (i.e. 1/d if time is in d).

    Returns
    -------
    numpy.ndarray[Any, dtype[float]]
        The noise spectrum calculated as the mean in a frequency window
        in the residual periodogram

    Notes
    -----
    The values calculated here capture the amount of noise on fitting a
    sinusoid of a certain frequency to all data points.
    Not to be confused with the noise on the individual data points of the
    time series.
    """
    # use defaults to get full amplitude spectrum
    freqs, ampls = scargle_parallel(time, resid)

    # determine the number of points to extend the spectrum with for convolution
    n_points = int(np.ceil(window_width / np.abs(freqs[1] - freqs[0])))  # .astype(int)
    window = np.full(n_points, 1 / n_points)

    # extend the array with mirrors for convolution
    ext_ampls = np.concatenate((ampls[(n_points - 1)::-1], ampls, ampls[:-(n_points + 1):-1]))
    ext_noise = np.convolve(ext_ampls, window, 'same')

    # cut back to original interval
    noise = ext_noise[n_points:-n_points]

    # extra correction to account for convolve mode='full' instead of 'same' (needed for JIT-ting)
    # noise = noise[n_points//2 - 1:-n_points//2]

    return noise


def scargle_noise_spectrum_redux(freqs, ampls, window_width=1.0):
    """Calculate the Lomb-Scargle noise spectrum by a convolution with a flat window of a certain width,
    given an amplitude spectrum.

    Parameters
    ----------
    freqs: numpy.ndarray[Any, dtype[float]]
        Frequencies at which the periodogram was calculated
    ampls: numpy.ndarray[Any, dtype[float]]
        The periodogram spectrum in the chosen units
    window_width: float
        The width of the window used to compute the noise spectrum,
        in inverse unit of the time array (i.e. 1/d if time is in d).

    Returns
    -------
    numpy.ndarray[Any, dtype[float]]
        The noise spectrum calculated as the mean in a frequency window
        in the residual periodogram

    Notes
    -----
    The values calculated here capture the amount of noise on fitting a
    sinusoid of a certain frequency to all data points.
    Not to be confused with the noise on the individual data points of the
    time series.
    """
    # determine the number of points to extend the spectrum with for convolution
    n_points = int(np.ceil(window_width / np.abs(freqs[1] - freqs[0])))  # .astype(int)
    window = np.full(n_points, 1 / n_points)

    # extend the array with mirrors for convolution
    ext_ampls = np.concatenate((ampls[(n_points - 1)::-1], ampls, ampls[:-(n_points + 1):-1]))
    ext_noise = np.convolve(ext_ampls, window, 'same')

    # cut back to original interval
    noise = ext_noise[n_points:-n_points]

    # extra correction to account for convolve mode='full' instead of 'same' (needed for JIT-ting)
    # noise = noise[n_points//2 - 1:-n_points//2]

    return noise


def scargle_noise_at_freq(fs, time, resid, window_width=1.):
    """Calculate the Lomb-Scargle noise at a given set of frequencies

    Parameters
    ----------
    fs: numpy.ndarray[Any, dtype[float]]
        The frequencies at which to calculate the noise
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    resid: numpy.ndarray[Any, dtype[float]]
        Residual measurement values of the time series
    window_width: float
        The width of the window used to compute the noise spectrum,
        in inverse unit of the time array (i.e. 1/d if time is in d).

    Returns
    -------
    numpy.ndarray[Any, dtype[float]]
        The noise level calculated as the mean in a window around the
        frequency in the residual periodogram

    Notes
    -----
    The values calculated here capture the amount of noise on fitting a
    sinusoid of a certain frequency to all data points.
    Not to be confused with the noise on the individual data points of the
    time series.
    """
    # use defaults to get full amplitude spectrum
    freqs, ampls = scargle_parallel(time, resid)
    margin = window_width / 2

    # mask the frequency ranges and compute the noise
    f_masks = [(freqs > f - margin) & (freqs <= f + margin) for f in fs]
    noise = np.array([np.mean(ampls[mask]) for mask in f_masks])

    return noise


def spectral_window(time, freqs):
    """Computes the modulus square of the spectral window W_N(f) of a set of
    time points at the given frequencies.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    freqs: numpy.ndarray[Any, dtype[float]]
        Frequency points to calculate the window. Inverse unit of `time`

    Returns
    -------
    numpy.ndarray[Any, dtype[float]]
        The spectral window at the given frequencies, |W(freqs)|^2

    Notes
    -----
    The spectral window is the Fourier transform of the window function
    w_N(t) = 1/N sum(Dirac(t - t_i))
    The time points do not need to be equidistant.
    The normalisation is such that 1.0 is returned at frequency 0.
    """
    n_time = len(time)
    cos_term = np.sum(np.cos(2.0 * np.pi * freqs * time.reshape(n_time, 1)), axis=0)
    sin_term = np.sum(np.sin(2.0 * np.pi * freqs * time.reshape(n_time, 1)), axis=0)
    win_kernel = cos_term ** 2 + sin_term ** 2

    # Normalise such that win_kernel(nu = 0.0) = 1.0
    spec_win = win_kernel / n_time ** 2

    return spec_win


@nb.njit(cache=True)
def _scargle_core(time, flux, nt, f0, df, nf):
    """Core algorithm of the Scargle periodogram with no weights.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series, mean subtracted.
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series, mean subtracted.
    nt: int
        Length of the time series.
    f0: float
        Starting frequency of the periodogram.
    df: float
        Frequency sampling space of the periodogram.
    nf: int
        Length of the frequency array.

    Returns
    -------
    numpy.ndarray[Any, dtype[float]]
        The periodogram spectrum in the chosen units.

    Notes
    -----
    Translated from Fortran (and just as fast when JIT-ted with Numba!)
        Computation of Scargles periodogram without explicit tau
        calculation, with iteration (Method Cuypers)

    Useful extra information: VanderPlas 2018,
    https://ui.adsabs.harvard.edu/abs/2018ApJS..236...16V/abstract
    """
    # pre-assign some memory
    ss = np.zeros(nf)
    sc = np.zeros(nf)
    ss2 = np.zeros(nf)
    sc2 = np.zeros(nf)

    # here is the actual calculation:
    two_pi = 2 * np.pi
    for i in range(nt):
        t_f0 = (time[i] * two_pi * f0) % two_pi
        sin_f0 = np.sin(t_f0)
        cos_f0 = np.cos(t_f0)
        mc_1_a = 2 * sin_f0 * cos_f0
        mc_1_b = cos_f0 * cos_f0 - sin_f0 * sin_f0

        t_df = (time[i] * two_pi * df) % two_pi
        sin_df = np.sin(t_df)
        cos_df = np.cos(t_df)
        mc_2_a = 2 * sin_df * cos_df
        mc_2_b = cos_df * cos_df - sin_df * sin_df

        sin_f0_s = sin_f0 * flux[i]
        cos_f0_s = cos_f0 * flux[i]
        for j in range(nf):
            ss[j] = ss[j] + sin_f0_s
            sc[j] = sc[j] + cos_f0_s
            temp_cos_f0_s = cos_f0_s
            cos_f0_s = temp_cos_f0_s * cos_df - sin_f0_s * sin_df
            sin_f0_s = sin_f0_s * cos_df + temp_cos_f0_s * sin_df

            ss2[j] = ss2[j] + mc_1_a
            sc2[j] = sc2[j] + mc_1_b
            temp_mc_1_b = mc_1_b
            mc_1_b = temp_mc_1_b * mc_2_b - mc_1_a * mc_2_a
            mc_1_a = mc_1_a * mc_2_b + temp_mc_1_b * mc_2_a

    s1 = ((sc ** 2 * (nt - sc2) + ss ** 2 * (nt + sc2) - 2 * ss * sc * ss2) / (nt ** 2 - sc2 ** 2 - ss2 ** 2))

    return s1


@nb.njit(cache=True)
def scargle(time, flux, f0=-1, fn=-1, df=-1, norm='amplitude'):
    """Scargle periodogram with no weights.

    The time array is mean subtracted to reduce correlation between frequencies and phases.
    The flux array is mean subtracted to avoid a large peak at frequency equal to zero.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    f0: float, optional
        Starting frequency of the periodogram.
        If left -1, default is f0 = 1/(100*T)
    fn: float, optional
        Last frequency of the periodogram.
        If left -1, default is fn = 1/(2*np.min(np.diff(time))) = Nyquist frequency
    df: float, optional
        Frequency sampling space of the periodogram
        If left -1, default is df = 1/(10*T) = oversampling factor of ten (recommended)
    norm: str, optional
        Normalisation of the periodogram. Choose from:
        'amplitude', 'density' or 'distribution'

    Returns
    -------
    tuple
        A tuple containing the following elements:
        f1: numpy.ndarray[Any, dtype[float]]
            Frequencies at which the periodogram was calculated
        s1: numpy.ndarray[Any, dtype[float]]
            The periodogram spectrum in the chosen units

    Notes
    -----
    Translated from Fortran (and just as fast when JIT-ted with Numba!)
        Computation of Scargles periodogram without explicit tau
        calculation, with iteration (Method Cuypers)

    Useful extra information: VanderPlas 2018,
    https://ui.adsabs.harvard.edu/abs/2018ApJS..236...16V/abstract
    """
    # time and flux are mean subtracted (reduce correlation and avoid peak at f=0)
    mean_t = np.mean(time)
    mean_s = np.mean(flux)
    time_sorter = np.argsort(time)
    time_ms = time[time_sorter] - mean_t
    flux_ms = flux[time_sorter] - mean_s

    # setup
    nt = len(time_ms)
    t_tot = np.ptp(time_ms)
    if f0 == -1:
        f0 = 0.01 / t_tot  # lower than T/100 no good
    if df == -1:
        df = 0.1 / t_tot  # default frequency sampling is about 1/10 of frequency resolution
    if fn == -1:
        fn = 1 / (2 * np.min(time_ms[1:] - time_ms[:-1]))
    nf = int((fn - f0) / df + 0.001) + 1

    # do the scargle calculations
    s1 = _scargle_core(time_ms, flux_ms, nt, f0, df, nf)
    f1 = f0 + np.arange(nf) * df

    # conversion to amplitude spectrum (or power density or statistical distribution)
    if not np.isfinite(s1[0]):
        s1[0] = 0  # sometimes there can be a nan value

    # convert to the wanted normalisation
    if norm == 'distribution':  # statistical distribution
        s1 /= np.var(flux_ms)
    elif norm == 'amplitude':  # amplitude spectrum
        s1 = np.sqrt(4 / nt) * np.sqrt(s1)
    elif norm == 'density':  # power density
        s1 = (4 / nt) * s1 * t_tot
    else:  # unnormalised (PSD?)
        s1 = s1

    return f1, s1


@nb.njit(cache=True, parallel=True)
def scargle_parallel(time, flux, f0=-1, fn=-1, df=-1, norm='amplitude'):
    """Parallel Scargle periodogram with no weights.

    Non-parallel overhead amounts to less than 10%.

    The time array is mean subtracted to reduce correlation between frequencies and phases.
    The flux array is mean subtracted to avoid a large peak at frequency equal to zero.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    f0: float, optional
        Starting frequency of the periodogram.
        If left -1, default is f0 = 1/(100*T)
    fn: float, optional
        Last frequency of the periodogram.
        If left -1, default is fn = 1/(2*np.min(np.diff(time))) = Nyquist frequency
    df: float, optional
        Frequency sampling space of the periodogram
        If left -1, default is df = 1/(10*T) = oversampling factor of ten (recommended)
    norm: str, optional
        Normalisation of the periodogram. Choose from:
        'amplitude', 'density' or 'distribution'

    Returns
    -------
    tuple
        A tuple containing the following elements:
        f1: numpy.ndarray[Any, dtype[float]]
            Frequencies at which the periodogram was calculated
        s1: numpy.ndarray[Any, dtype[float]]
            The periodogram spectrum in the chosen units

    Notes
    -----
    Translated from Fortran (and just as fast when JIT-ted with Numba!)
        Computation of Scargles periodogram without explicit tau
        calculation, with iteration (Method Cuypers)

    Useful extra information: VanderPlas 2018,
    https://ui.adsabs.harvard.edu/abs/2018ApJS..236...16V/abstract
    """
    # time and flux are mean subtracted (reduce correlation and avoid peak at f=0)
    mean_t = np.mean(time)
    mean_s = np.mean(flux)
    time_sorter = np.argsort(time)
    time_ms = time[time_sorter] - mean_t
    flux_ms = flux[time_sorter] - mean_s

    # setup
    nt = len(time_ms)
    t_tot = np.ptp(time_ms)
    if f0 == -1:
        f0 = 0.01 / t_tot  # lower than T/100 no good
    if df == -1:
        df = 0.1 / t_tot  # default frequency sampling is about 1/10 of frequency resolution
    if fn == -1:
        fn = 1 / (2 * np.min(time_ms[1:] - time_ms[:-1]))
    nf = int((fn - f0) / df + 0.001) + 1

    # frequency array, broken into chunks
    f1 = f0 + np.arange(nf) * df
    f1_chunks = np.array_split(f1, n_proc)
    chunk_i = np.array([len(chunk) for chunk in f1_chunks])
    chunk_i = np.array([np.sum(chunk_i[:i]) for i in range(1, len(chunk_i) + 1)])
    chunk_i = np.concatenate((np.zeros(1), chunk_i)).astype(np.int_)

    # parallelised for loop
    s1 = np.zeros(nf)
    for i in nb.prange(n_proc):
        _f0 = f1_chunks[i][0]
        _nf = len(f1_chunks[i])
        s1[chunk_i[i]:chunk_i[i + 1]] = _scargle_core(time_ms, flux_ms, nt, _f0, df, _nf)

    # conversion to amplitude spectrum (or power density or statistical distribution)
    if not np.isfinite(s1[0]):
        s1[0] = 0  # sometimes there can be a nan value

    # convert to the wanted normalisation
    if norm == 'distribution':  # statistical distribution
        s1 /= np.var(flux_ms)
    elif norm == 'amplitude':  # amplitude spectrum
        s1 = np.sqrt(4 / nt) * np.sqrt(s1)
    elif norm == 'density':  # power density
        s1 = (4 / nt) * s1 * t_tot
    else:  # unnormalised (PSD?)
        s1 = s1

    return f1, s1


@nb.njit(cache=True)
def scargle_ampl_phase_single(time, flux, f):
    """Amplitude and phase at one or a set of frequencies from the Scargle periodogram.

    The time array is mean subtracted to reduce correlation between frequencies and phases.
    The flux array is mean subtracted to avoid a large peak at frequency equal to zero.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series.
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series.
    f: float
        A frequency to calculate amplitude and phase at.

    Returns
    -------
    tuple
        Two numbers consisting of:
        float
            Amplitude at the given frequency.
        float
            Phase at the given frequency.

    See Also
    --------
    scargle_phase

    Notes
    -----
    For the phase calculation:
    Uses a slightly modified version of the function in Hocke 1997
    ("Phase estimation with the Lomb-Scargle periodogram method")
    https://www.researchgate.net/publication/283359043_Phase_estimation_with_the_Lomb-Scargle_periodogram_method
    (only difference is an extra pi/2 for changing cos phase to sin phase)
    """
    # time and flux are mean subtracted (reduce correlation and avoid peak at f=0)
    mean_t = np.mean(time)
    mean_s = np.mean(flux)
    time_ms = time - mean_t
    flux_ms = flux - mean_s

    # setup
    nt = len(time_ms)
    pi = np.pi
    two_pi = 2 * pi
    four_pi = 4 * pi

    # define tau
    cos_tau = 0
    sin_tau = 0
    for j in range(nt):
        cos_tau += np.cos(four_pi * f * time_ms[j])
        sin_tau += np.sin(four_pi * f * time_ms[j])
    tau = 1 / (four_pi * f) * np.arctan2(sin_tau, cos_tau)  # tau(f)

    # define the general cos and sin functions
    s_cos = 0
    cos_2 = 0
    s_sin = 0
    sin_2 = 0
    for j in range(nt):
        cos = np.cos(two_pi * f * (time_ms[j] - tau))
        sin = np.sin(two_pi * f * (time_ms[j] - tau))
        s_cos += flux_ms[j] * cos
        cos_2 += cos ** 2
        s_sin += flux_ms[j] * sin
        sin_2 += sin ** 2

    # final calculations
    a_cos = s_cos / cos_2 ** (1 / 2)
    b_sin = s_sin / sin_2 ** (1 / 2)

    # amplitude
    ampl = (a_cos ** 2 + b_sin ** 2) / 2
    ampl = np.sqrt(4 / nt) * np.sqrt(ampl)  # conversion to amplitude

    # sine phase (radians)
    phi = pi / 2 - np.arctan2(b_sin, a_cos) - two_pi * f * tau
    phi = (phi + pi) % two_pi - pi  # make sure the phase stays within + and - pi

    return ampl, phi


@nb.njit(cache=True, parallel=True)
def scargle_ampl_phase(time, flux, fs):
    """Amplitude at one or a set of frequencies from the Scargle periodogram.

    Fast for limited numbers of frequencies, not a replacement for the full periodogram.
    Has an overhead compared to scargle_ampl_phase_single, so do not use for single fs value.
    For 2 frequencies it is just as fast, and 3+ it is more efficient.

    The time array is mean subtracted to reduce correlation between frequencies and phases.
    The flux array is mean subtracted to avoid a large peak at frequency equal to zero.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series.
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series.
    fs: numpy.ndarray[Any, dtype[float]]
        A set of frequencies to calculate amplitude and phase at.

    Returns
    -------
    tuple
        Two arrays consisting of:
        numpy.ndarray[Any, dtype[float]]
            Amplitude at the given frequencies
        numpy.ndarray[Any, dtype[float]]
            Phase at the given frequencies

    See Also
    --------
    scargle_phase

    Notes
    -----
    For the phase calculation:
    Uses a slightly modified version of the function in Hocke 1997
    ("Phase estimation with the Lomb-Scargle periodogram method")
    https://www.researchgate.net/publication/283359043_Phase_estimation_with_the_Lomb-Scargle_periodogram_method
    (only difference is an extra pi/2 for changing cos phase to sin phase)
    """
    # time and flux are mean subtracted (reduce correlation and avoid peak at f=0)
    mean_t = np.mean(time)
    mean_s = np.mean(flux)
    time_ms = time - mean_t
    flux_ms = flux - mean_s

    # setup
    nt = len(time_ms)
    pi = np.pi
    two_pi = 2 * pi
    four_pi = 4 * pi
    fs = np.atleast_1d(fs)
    ampl = np.zeros(len(fs))
    phi = np.zeros(len(fs))

    for i in nb.prange(len(fs)):
        # define tau
        cos_tau = 0
        sin_tau = 0
        for j in range(nt):
            cos_tau += np.cos(four_pi * fs[i] * time_ms[j])
            sin_tau += np.sin(four_pi * fs[i] * time_ms[j])
        tau = 1 / (four_pi * fs[i]) * np.arctan2(sin_tau, cos_tau)  # tau(f)

        # define the general cos and sin functions
        s_cos = 0
        cos_2 = 0
        s_sin = 0
        sin_2 = 0
        for j in range(nt):
            cos = np.cos(two_pi * fs[i] * (time_ms[j] - tau))
            sin = np.sin(two_pi * fs[i] * (time_ms[j] - tau))
            s_cos += flux_ms[j] * cos
            cos_2 += cos ** 2
            s_sin += flux_ms[j] * sin
            sin_2 += sin ** 2

        # final calculations
        a_cos = s_cos / cos_2 ** (1 / 2)
        b_sin = s_sin / sin_2 ** (1 / 2)

        # amplitude
        ampl[i] = (a_cos ** 2 + b_sin ** 2) / 2
        ampl[i] = np.sqrt(4 / nt) * np.sqrt(ampl[i])  # conversion to amplitude

        # sine phase (radians)
        phi[i] = pi / 2 - np.arctan2(b_sin, a_cos) - two_pi * fs[i] * tau
        phi[i] = (phi[i] + pi) % two_pi - pi  # make sure the phase stays within + and - pi

    return ampl, phi


def astropy_scargle(time, flux, f0=0, fn=0, df=0, norm='amplitude'):
    """Wrapper for the astropy Scargle periodogram.

    The time array is mean subtracted to reduce correlation between frequencies and phases.
    The flux array is mean subtracted to avoid a large peak at frequency equal to zero.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    f0: float
        Starting frequency of the periodogram.
        If left zero, default is f0 = 1/(100*T)
    fn: float
        Last frequency of the periodogram.
        If left zero, default is fn = 1/(2*np.min(np.diff(time))) = Nyquist frequency
    df: float
        Frequency sampling space of the periodogram
        If left zero, default is df = 1/(10*T) = oversampling factor of ten (recommended)
    norm: str
        Normalisation of the periodogram. Choose from:
        'amplitude', 'density' or 'distribution'

    Returns
    -------
    tuple
        A tuple containing the following elements:
        f1: numpy.ndarray[Any, dtype[float]]
            Frequencies at which the periodogram was calculated
        s1: numpy.ndarray[Any, dtype[float]]
            The periodogram spectrum in the chosen units

    Notes
    -----
    Approximation using fft, much faster (in mode='fast') than the single threaded scargle (about x10).
    Note that the astropy implementation uses functions under the hood that use the blas package for
    multithreading by default. Compared to the parallel scargle it is similar in speed.
    Beware of computing narrower frequency windows, as there is inconsistency when doing this.
    It is also generally less accurate.

    Useful extra information: VanderPlas 2018,
    https://ui.adsabs.harvard.edu/abs/2018ApJS..236...16V/abstract
    """
    # time and flux are mean subtracted (reduce correlation and avoid peak at f=0)
    mean_t = np.mean(time)
    mean_s = np.mean(flux)
    time_sorter = np.argsort(time)
    time_ms = time[time_sorter] - mean_t
    flux_ms = flux[time_sorter] - mean_s

    # setup
    n = len(flux)
    t_tot = np.ptp(time_ms)
    f0 = max(f0, 0.01 / t_tot)  # don't go lower than T/100
    if df == 0:
        df = 0.1 / t_tot  # default frequency sampling is about 1/10 of frequency resolution
    if fn == 0:
        fn = 1 / (2 * np.min(time_ms[1:] - time_ms[:-1]))
    nf = int((fn - f0) / df + 0.001) + 1
    f1 = f0 + np.arange(nf) * df

    # use the astropy fast algorithm and normalise afterward
    ls = apy.LombScargle(time_ms, flux_ms, fit_mean=False, center_data=False)
    s1 = ls.power(f1, normalization='psd', method='fast', assume_regular_frequency=True)

    # replace negative by zero (just in case - have seen it happen)
    s1[s1 < 0] = 0

    # convert to the wanted normalisation
    if norm == 'distribution':  # statistical distribution
        s1 /= np.var(flux_ms)
    elif norm == 'amplitude':  # amplitude spectrum
        s1 = np.sqrt(4 / n) * np.sqrt(s1)
    elif norm == 'density':  # power density
        s1 = (4 / n) * s1 * t_tot
    else:  # unnormalised (PSD?)
        s1 = s1

    return f1, s1



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains functions for visualisation;
specifically for visualising the analysis of stellar variability and harmonic sinusoids.

Code written by: Luc IJspeert
"""

import os

import numpy as np
import matplotlib.pyplot as plt

try:
    import arviz as az  # optional functionality
except ImportError:
    az = None
    pass

from star_shine.core import time_series as tms, model as mdl, frequency_sets as frs
from star_shine.core import periodogram as pdg, utility as ut, io
from star_shine.config.helpers import get_mpl_stylesheet_path


# mpl style sheet
plt.style.use(get_mpl_stylesheet_path())


def plot_pd(time, flux, i_chunks, plot_per_chunk=False, file_name=None, show=True):
    """Plot the periodogram of the data.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
    plot_per_chunk: bool
        If True, plot the periodogram of all time chunks in one plot.
    file_name: str, optional
        File path to save the plot
    show: bool, optional
        If True, display the plot

    Returns
    -------
    None
    """
    # make the periodograms
    freqs, ampls = pdg.scargle_parallel(time, flux)
    freqs_list, ampls_list = [], []
    if plot_per_chunk:
        for ch in i_chunks:
            f, a = pdg.scargle_parallel(time[ch[0]:ch[1]], flux[ch[0]:ch[1]])
            freqs_list.append(f)
            ampls_list.append(a)

    # plot ranges
    freq_range = np.ptp(freqs)

    # plot
    fig, ax = plt.subplots()

    if plot_per_chunk:
        for i in range(len(i_chunks)):
            ax.plot(freqs_list[i], ampls_list[i])
    else:
        ax.plot(freqs, ampls, c='tab:blue', label='flux')

    ax.set_xlim(freqs[0] - freq_range * 0.05, freqs[-1] + freq_range * 0.05)
    plt.xlabel('frequency (1/d)')
    plt.ylabel('amplitude')
    plt.legend()
    plt.tight_layout()

    # save/show/close
    if file_name is not None:
        plt.savefig(file_name, dpi=120, format='png')  # 16 by 9 at 120 dpi is 1080p
    if show:
        plt.show()
    else:
        plt.close()

    return None

def plot_pd_single_output(time, flux, flux_err, p_orb, p_err, const, slope, f_n, a_n, ph_n, i_chunks,
                          annotate=True, file_name=None, show=True):
    """Plot the periodogram with one output of the analysis recipe.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    flux_err: numpy.ndarray[Any, dtype[float]]
        Errors in the measurement values
    p_orb: float
        Orbital period
    p_err: float
        Error associated with the orbital period
    const: numpy.ndarray[Any, dtype[float]]
        The y-intercepts of a piece-wise linear curve
    slope: numpy.ndarray[Any, dtype[float]]
        The slopes of a piece-wise linear curve
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sine waves
    ph_n: numpy.ndarray[Any, dtype[float]]
        The phases of a number of sine waves
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
    annotate: bool, optional
        If True, annotate the plot with the frequencies.
    file_name: str, optional
        File path to save the plot
    show: bool, optional
        If True, display the plot

    Returns
    -------
    None
    """
    # separate harmonics
    harmonics, harmonic_n = frs.find_harmonics_from_pattern(f_n, 1/p_orb, f_tol=1e-9)

    # make model
    model_linear = mdl.linear_curve(time, const, slope, i_chunks)
    model_sinusoid = mdl.sum_sines(time, f_n, a_n, ph_n)
    model = model_linear + model_sinusoid

    # make periodograms
    freqs, ampls = pdg.scargle_parallel(time, flux)
    freq_range = np.ptp(freqs)
    freqs_r, ampls_r = pdg.scargle_parallel(time, flux - model)

    # get error values
    errors = ut.formal_uncertainties(time, flux - model, flux_err, a_n, i_chunks)

    # max plot value
    y_max = max(np.max(ampls), np.max(a_n))

    # plot
    fig, ax = plt.subplots()
    if len(harmonics) > 0:
        ax.errorbar([1 / p_orb, 1 / p_orb], [0, y_max], xerr=[0, p_err / p_orb**2],
                    linestyle='-', capsize=2, c='tab:grey', label=f'orbital frequency (p={p_orb:1.4f}d +-{p_err:1.4f})')
        for i in range(2, np.max(harmonic_n) + 1):
            ax.plot([i / p_orb, i / p_orb], [0, y_max], linestyle='-', c='tab:grey', alpha=0.3)
        ax.errorbar([], [], xerr=[], yerr=[], linestyle='-', capsize=2, c='tab:red', label='extracted harmonics')
    ax.plot(freqs, ampls, c='tab:blue', label='flux')
    ax.plot(freqs_r, ampls_r, c='tab:orange', label='residual')
    for i in range(len(f_n)):
        if i in harmonics:
            ax.errorbar([f_n[i], f_n[i]], [0, a_n[i]], xerr=[0, errors[2][i]], yerr=[0, errors[3][i]],
                        linestyle='-', capsize=2, c='tab:red')
        else:
            ax.errorbar([f_n[i], f_n[i]], [0, a_n[i]], xerr=[0, errors[2][i]], yerr=[0, errors[3][i]],
                        linestyle='-', capsize=2, c='tab:pink')
        if annotate:
            ax.annotate(f'{i + 1}', (f_n[i], a_n[i]))
    ax.errorbar([], [], xerr=[], yerr=[], linestyle='-', capsize=2, c='tab:pink', label='extracted frequencies')
    ax.set_xlim(freqs[0] - freq_range * 0.05, freqs[-1] + freq_range * 0.05)
    plt.xlabel('frequency (1/d)')
    plt.ylabel('amplitude')
    plt.legend()
    plt.tight_layout()

    # save/show/close
    if file_name is not None:
        plt.savefig(file_name, dpi=120, format='png')  # 16 by 9 at 120 dpi is 1080p
    if show:
        plt.show()
    else:
        plt.close()

    return None


def plot_pd_full_output(time, flux, flux_err, models, p_orb_i, p_err_i, f_n_i, a_n_i, i_chunks, file_name=None,
                        show=True):
    """Plot the periodogram with the full output of the analysis recipe.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    flux_err: numpy.ndarray[Any, dtype[float]]
        Errors in the measurement values
    models: list[numpy.ndarray[Any, dtype[float]]]
        List of model fluxs for different stages of the analysis
    p_orb_i: list[float]
        Orbital periods for different stages of the analysis
    p_err_i: list[float]
        Errors associated with the orbital periods
        for different stages of the analysis
    f_n_i: list[numpy.ndarray[Any, dtype[float]]]
        List of extracted frequencies for different stages of the analysis
    a_n_i: list[numpy.ndarray[Any, dtype[float]]]
        List of amplitudes corresponding to the extracted frequencies
        for different stages of the analysis
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
    file_name: str, optional
        File path to save the plot
    show: bool, optional
        If True, display the plot

    Returns
    -------
    None
    """
    # make periodograms
    freqs, ampls = pdg.scargle_parallel(time, flux - np.mean(flux))
    freq_range = np.ptp(freqs)
    freqs_1, ampls_1 = pdg.scargle_parallel(time, flux - models[0] - np.all(models[0] == 0) * np.mean(flux))
    freqs_2, ampls_2 = pdg.scargle_parallel(time, flux - models[1] - np.all(models[1] == 0) * np.mean(flux))
    freqs_3, ampls_3 = pdg.scargle_parallel(time, flux - models[2] - np.all(models[2] == 0) * np.mean(flux))
    freqs_4, ampls_4 = pdg.scargle_parallel(time, flux - models[3] - np.all(models[3] == 0) * np.mean(flux))
    freqs_5, ampls_5 = pdg.scargle_parallel(time, flux - models[4] - np.all(models[4] == 0) * np.mean(flux))

    # get error values
    err_1 = ut.formal_uncertainties(time, flux - models[0], flux_err, a_n_i[0], i_chunks)
    err_2 = ut.formal_uncertainties(time, flux - models[1], flux_err, a_n_i[1], i_chunks)
    err_3 = ut.formal_uncertainties(time, flux - models[2], flux_err, a_n_i[2], i_chunks)
    err_4 = ut.formal_uncertainties(time, flux - models[3], flux_err, a_n_i[3], i_chunks)
    err_5 = ut.formal_uncertainties(time, flux - models[4], flux_err, a_n_i[4], i_chunks)

    # max plot value
    if len(f_n_i[4]) > 0:
        y_max = max(np.max(ampls), np.max(a_n_i[4]))
    else:
        y_max = np.max(ampls)

    # plot
    fig, ax = plt.subplots()
    ax.plot(freqs, ampls, label='flux')
    if len(f_n_i[0]) > 0:
        ax.plot(freqs_1, ampls_1, label='extraction residual')
    if len(f_n_i[1]) > 0:
        ax.plot(freqs_2, ampls_2, label='NL-LS optimisation residual')
    if len(f_n_i[2]) > 0:
        ax.plot(freqs_3, ampls_3, label='coupled harmonics residual')
    if len(f_n_i[3]) > 0:
        ax.plot(freqs_4, ampls_4, label='additional frequencies residual')
    if len(f_n_i[4]) > 0:
        ax.plot(freqs_5, ampls_5, label='NL-LS fit residual with harmonics residual')

    # period
    if p_orb_i[4] > 0:
        ax.errorbar([1 / p_orb_i[4], 1 / p_orb_i[4]], [0, y_max], xerr=[0, p_err_i[4] / p_orb_i[4]**2],
                    linestyle='--', capsize=2, c='k', label=f'orbital frequency (p={p_orb_i[4]:1.4f}d)')
    elif p_orb_i[2] > 0:
        ax.errorbar([1 / p_orb_i[2], 1 / p_orb_i[2]], [0, y_max], xerr=[0, p_err_i[2] / p_orb_i[2]**2],
                    linestyle='--', capsize=2, c='k', label=f'orbital frequency (p={p_orb_i[2]:1.4f}d)')

    # frequencies
    for i in range(len(f_n_i[0])):
        ax.errorbar([f_n_i[0][i], f_n_i[0][i]], [0, a_n_i[0][i]], xerr=[0, err_1[2][i]], yerr=[0, err_1[3][i]],
                    linestyle=':', capsize=2, c='tab:orange')
    for i in range(len(f_n_i[1])):
        ax.errorbar([f_n_i[1][i], f_n_i[1][i]], [0, a_n_i[1][i]], xerr=[0, err_2[2][i]], yerr=[0, err_2[3][i]],
                    linestyle=':', capsize=2, c='tab:green')
    for i in range(len(f_n_i[2])):
        ax.errorbar([f_n_i[2][i], f_n_i[2][i]], [0, a_n_i[2][i]], xerr=[0, err_3[2][i]], yerr=[0, err_3[3][i]],
                    linestyle=':', capsize=2, c='tab:red')
    for i in range(len(f_n_i[3])):
        ax.errorbar([f_n_i[3][i], f_n_i[3][i]], [0, a_n_i[3][i]], xerr=[0, err_4[2][i]], yerr=[0, err_4[3][i]],
                    linestyle=':', capsize=2, c='tab:purple')
    for i in range(len(f_n_i[4])):
        ax.errorbar([f_n_i[4][i], f_n_i[4][i]], [0, a_n_i[4][i]], xerr=[0, err_5[2][i]], yerr=[0, err_5[3][i]],
                    linestyle=':', capsize=2, c='tab:brown')
        ax.annotate(f'{i + 1}', (f_n_i[4][i], a_n_i[4][i]))
    ax.set_xlim(freqs[0] - freq_range * 0.05, freqs[-1] + freq_range * 0.05)
    plt.xlabel('frequency (1/d)')
    plt.ylabel('amplitude')
    plt.legend()
    plt.tight_layout()

    # save/show/close
    if file_name is not None:
        plt.savefig(file_name, dpi=120, format='png')  # 16 by 9 at 120 dpi is 1080p
    if show:
        plt.show()
    else:
        plt.close()

    return None


def plot_lc(time, flux, flux_err, i_chunks, file_name=None, show=True):
    """Shows the light curve data

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    flux_err: numpy.ndarray[Any, dtype[float]]
        Errors in the measurement values
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
    file_name: str, optional
        File path to save the plot
    show: bool, optional
        If True, display the plot

    Returns
    -------
    None
    """
    # plot the light curve data with different colours for each chunk
    fig, ax = plt.subplots(figsize=(16, 9))
    for ch in i_chunks:
        t_mean = np.mean(time[ch[0], ch[1]])
        f_min = np.min(flux[ch[0], ch[1]])
        f_max = np.max(flux[ch[0], ch[1]])
        ax.plot([t_mean, t_mean], [f_min, f_max], alpha=0.3)
        ax.errorbar(time[ch[0], ch[1]], flux[ch[0], ch[1]], yerr=flux_err[ch[0], ch[1]], color='grey', alpha=0.3)
        ax.scatter(time[ch[0], ch[1]], flux[ch[0], ch[1]], marker='.', label='dataset')
    ax.set_xlabel('time')
    ax.set_ylabel('flux')
    ax.legend()
    plt.tight_layout()

    # save/show/close
    if file_name is not None:
        plt.savefig(file_name, dpi=120, format='png')  # 16 by 9 at 120 dpi is 1080p
    if show:
        plt.show()
    else:
        plt.close()

    return None


def plot_lc_sinusoids(time, flux, const, slope, f_n, a_n, ph_n, i_chunks, file_name=None, show=True):
    """Shows the separated harmonics in several ways

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    const: numpy.ndarray[Any, dtype[float]]
        The y-intercepts of a piece-wise linear curve
    slope: numpy.ndarray[Any, dtype[float]]
        The slopes of a piece-wise linear curve
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sine waves
    ph_n: numpy.ndarray[Any, dtype[float]]
        The phases of a number of sine waves
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
    file_name: str, optional
        File path to save the plot
    show: bool, optional
        If True, display the plot

    Returns
    -------
    None
    """
    t_mean = np.mean(time)

    # make models
    model_linear = mdl.linear_curve(time, const, slope, i_chunks)
    model_sines = mdl.sum_sines(time, f_n, a_n, ph_n)
    resid = flux - (model_linear + model_sines)

    # plot the full model light curve
    fig, ax = plt.subplots(nrows=2, sharex=True)
    ax[0].plot([t_mean, t_mean], [np.min(flux), np.max(flux)], c='grey', alpha=0.3)
    ax[0].scatter(time, flux, marker='.', label='flux')
    ax[0].plot(time, model_linear + model_sines, c='tab:orange', label='full model (linear + sinusoidal)')
    ax[1].plot([t_mean, t_mean], [np.min(resid), np.max(resid)], c='grey', alpha=0.3)
    ax[1].scatter(time, resid, marker='.')
    ax[0].set_ylabel('flux/model')
    ax[0].legend()
    ax[1].set_ylabel('residual')
    ax[1].set_xlabel('time (d)')
    plt.tight_layout()
    plt.subplots_adjust(wspace=0, hspace=0)

    # save/show/close
    if file_name is not None:
        plt.savefig(file_name, dpi=120, format='png')  # 16 by 9 at 120 dpi is 1080p
    if show:
        plt.show()
    else:
        plt.close()

    return None


def plot_lc_harmonics(time, flux, p_orb, p_err, const, slope, f_n, a_n, ph_n, i_chunks, file_name=None,
                      show=True):
    """Shows the separated harmonics in several ways

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    p_orb: float
        Orbital period of the system
    p_err: float
        Error in the orbital period
    const: numpy.ndarray[Any, dtype[float]]
        The y-intercepts of a piece-wise linear curve
    slope: numpy.ndarray[Any, dtype[float]]
        The slopes of a piece-wise linear curve
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sine waves
    ph_n: numpy.ndarray[Any, dtype[float]]
        The phases of a number of sine waves
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
    file_name: str, optional
        File path to save the plot
    show: bool, optional
        If True, display the plot

    Returns
    -------
    None
    """
    t_mean = np.mean(time)

    # make models
    model_line = mdl.linear_curve(time, const, slope, i_chunks)
    harmonics, harmonic_n = frs.find_harmonics_from_pattern(f_n, 1/p_orb, f_tol=1e-9)
    model_h = mdl.sum_sines(time, f_n[harmonics], a_n[harmonics], ph_n[harmonics])
    model_nh = mdl.sum_sines(time, np.delete(f_n, harmonics), np.delete(a_n, harmonics),
                                                     np.delete(ph_n, harmonics))
    resid_nh = flux - model_nh
    resid_h = flux - model_h

    # plot the harmonic model and non-harmonic model
    fig, ax = plt.subplots(nrows=2, sharex=True)
    ax[0].plot([t_mean, t_mean], [np.min(resid_nh), np.max(resid_nh)], c='grey', alpha=0.3)
    ax[0].scatter(time, resid_nh, marker='.', c='tab:blue', label='flux - non-harmonics')
    ax[0].plot(time, model_line + model_h, c='tab:orange', label='linear + harmonic model, '
                                                                  f'p={p_orb:1.4f}d (+-{p_err:1.4f})')
    ax[1].plot([t_mean, t_mean], [np.min(resid_h), np.max(resid_h)], c='grey', alpha=0.3)
    ax[1].scatter(time, resid_h, marker='.', c='tab:blue', label='flux - harmonics')
    ax[1].plot(time, model_line + model_nh, c='tab:orange', label='linear + non-harmonic model')
    ax[0].set_ylabel('residual/model')
    ax[0].legend()
    ax[1].set_ylabel('residual/model')
    ax[1].set_xlabel('time (d)')
    ax[1].legend()
    plt.tight_layout()
    plt.subplots_adjust(wspace=0, hspace=0)

    # save/show/close
    if file_name is not None:
        plt.savefig(file_name, dpi=120, format='png')  # 16 by 9 at 120 dpi is 1080p
    if show:
        plt.show()
    else:
        plt.close()

    return


def plot_trace_sinusoids(inf_data, const, slope, f_n, a_n, ph_n):
    """Show the pymc3 sampling results in a trace plot

    Parameters
    ----------
    inf_data: InferenceData object
        Arviz inference data object
    const: numpy.ndarray[Any, dtype[float]]
        The y-intercepts of a piece-wise linear curve
    slope: numpy.ndarray[Any, dtype[float]]
        The slopes of a piece-wise linear curve
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sine waves
    ph_n: numpy.ndarray[Any, dtype[float]]
        The phases of a number of sine waves

    Returns
    -------
    None
    """
    # convert phases to interval [-pi, pi] from [0, 2pi]
    above_pi = (ph_n >= np.pi)
    ph_n[above_pi] = ph_n[above_pi] - 2 * np.pi
    par_lines = [('const', {}, const), ('slope', {}, slope), ('f_n', {}, f_n), ('a_n', {}, a_n), ('ph_n', {}, ph_n)]
    az.plot_trace(inf_data, combined=False, compact=True, rug=True, divergences='top', lines=par_lines)

    return


def plot_pair_harmonics(inf_data, p_orb, const, slope, f_n, a_n, ph_n, save_file=None, show=True):
    """Show the pymc3 sampling results in several pair plots

    Parameters
    ----------
    inf_data: object
        Arviz inference data object
    p_orb: float
        Orbital period
    const: numpy.ndarray[Any, dtype[float]]
        The y-intercepts of a piece-wise linear curve
    slope: numpy.ndarray[Any, dtype[float]]
        The slopes of a piece-wise linear curve
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sine waves
    ph_n: numpy.ndarray[Any, dtype[float]]
        The phases of a number of sine waves
    save_file: str, optional
        File path to save the plot
    show: bool, optional
        If True, display the plot

    Returns
    -------
    None
    """
    # convert phases to interval [-pi, pi] from [0, 2pi]
    above_pi = (ph_n >= np.pi)
    ph_n[above_pi] = ph_n[above_pi] - 2 * np.pi
    harmonics, harmonic_n = frs.find_harmonics_from_pattern(f_n, 1/p_orb, f_tol=1e-9)
    non_harm = np.delete(np.arange(len(f_n)), harmonics)
    ref_values = {'p_orb': p_orb, 'const': const, 'slope': slope,
                  'f_n': f_n[non_harm], 'a_n': a_n[non_harm], 'ph_n': ph_n[non_harm],
                  'f_h': f_n[harmonics], 'a_h': a_n[harmonics], 'ph_h': ph_n[harmonics]}
    kwargs = {'marginals': True, 'textsize': 14, 'kind': ['scatter', 'kde'],
              'marginal_kwargs': {'quantiles': [0.158, 0.5, 0.842]}, 'point_estimate': 'mean',
              'reference_values': ref_values, 'show': show}
    az.plot_pair(inf_data, var_names=['f_n', 'a_n', 'ph_n'],
                 coords={'f_n_dim_0': [0, 1, 2], 'a_n_dim_0': [0, 1, 2], 'ph_n_dim_0': [0, 1, 2]}, **kwargs)
    az.plot_pair(inf_data, var_names=['p_orb', 'f_n'], coords={'f_n_dim_0': np.arange(9)}, **kwargs)
    ax = az.plot_pair(inf_data, var_names=['p_orb', 'const', 'slope', 'f_n', 'a_n', 'ph_n', 'a_h', 'ph_h'],
                      coords={'const_dim_0': [0], 'slope_dim_0': [0], 'f_n_dim_0': [0], 'a_n_dim_0': [0],
                              'ph_n_dim_0': [0], 'a_h_dim_0': [0], 'ph_h_dim_0': [0]}, **kwargs)

    # save if wanted (only last plot - most interesting one)
    if save_file is not None:
        fig = ax.ravel()[0].figure
        fig.savefig(save_file, dpi=120, format='png')  # 16 by 9 at 120 dpi is 1080p

    return


def plot_trace_harmonics(inf_data, p_orb, const, slope, f_n, a_n, ph_n):
    """Show the pymc3 sampling results in a trace plot

    Parameters
    ----------
    inf_data: InferenceData object
        Arviz inference data object
    p_orb: float
        Orbital period
    const: numpy.ndarray[Any, dtype[float]]
        The y-intercepts of a piece-wise linear curve
    slope: numpy.ndarray[Any, dtype[float]]
        The slopes of a piece-wise linear curve
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sine waves
    ph_n: numpy.ndarray[Any, dtype[float]]
        The phases of a number of sine waves

    Returns
    -------
    None
    """
    # convert phases to interval [-pi, pi] from [0, 2pi]
    above_pi = (ph_n >= np.pi)
    ph_n[above_pi] = ph_n[above_pi] - 2 * np.pi
    harmonics, harmonic_n = frs.find_harmonics_from_pattern(f_n, 1/p_orb, f_tol=1e-9)
    non_harm = np.delete(np.arange(len(f_n)), harmonics)
    par_lines = [('p_orb', {}, p_orb), ('const', {}, const), ('slope', {}, slope),
                 ('f_n', {}, f_n[non_harm]), ('a_n', {}, a_n[non_harm]), ('ph_n', {}, ph_n[non_harm]),
                 ('f_h', {}, f_n[harmonics]), ('a_h', {}, a_n[harmonics]), ('ph_h', {}, ph_n[harmonics])]
    az.plot_trace(inf_data, combined=False, compact=True, rug=True, divergences='top', lines=par_lines)

    return


def sequential_plotting(time, flux, flux_err, i_chunks, target_id, load_dir, save_dir=None, show=True):
    """Due to plotting not working under multiprocessing this function is
    made to make plots after running the analysis in parallel.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    flux_err: numpy.ndarray[Any, dtype[float]]
        Errors in the measurement values
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
    target_id: int, str
        In case of using analyse_from_tic:
        The TESS Input Catalog number
        In case of user-defined light curve files (analyse_from_file):
        Should be the same as the name of the light curve file.
    load_dir: str
        Path to a directory for loading analysis results.
        Will append <target_id> + _analysis automatically
    save_dir: str, None
        Path to a directory for saving the plots.
        Will append <target_id> + _analysis automatically
        Directory is created if it doesn't exist yet
    show: bool
        Whether to show the plots or not.

    Returns
    -------
    None
    """
    load_dir = os.path.join(load_dir, f'{target_id}_analysis')  # add subdir
    if save_dir is not None:
        save_dir = os.path.join(save_dir, f'{target_id}_analysis')  # add subdir

        # for saving, make a folder if not there yet
        if not os.path.isdir(save_dir):
            os.mkdir(save_dir)  # create the subdir

    # open all the data
    file_name = os.path.join(load_dir, f'{target_id}_analysis_1.hdf5')
    if os.path.isfile(file_name):
        results = io.load_result_hdf5(file_name)
        const_1, slope_1, f_n_1, a_n_1, ph_n_1 = results['sin_mean']
        model_linear = mdl.linear_curve(time, const_1, slope_1, i_chunks)
        model_sinusoid = mdl.sum_sines(time, f_n_1, a_n_1, ph_n_1)
        model_1 = model_linear + model_sinusoid
    else:
        const_1, slope_1, f_n_1, a_n_1, ph_n_1 = np.array([[], [], [], [], []])
        model_1 = np.zeros(len(time))

    file_name = os.path.join(load_dir, f'{target_id}_analysis_2.hdf5')
    if os.path.isfile(file_name):
        results = io.load_result_hdf5(file_name)
        const_2, slope_2, f_n_2, a_n_2, ph_n_2 = results['sin_mean']
        model_linear = mdl.linear_curve(time, const_2, slope_2, i_chunks)
        model_sinusoid = mdl.sum_sines(time, f_n_2, a_n_2, ph_n_2)
        model_2 = model_linear + model_sinusoid
    else:
        const_2, slope_2, f_n_2, a_n_2, ph_n_2 = np.array([[], [], [], [], []])
        model_2 = np.zeros(len(time))

    file_name = os.path.join(load_dir, f'{target_id}_analysis_3.hdf5')
    if os.path.isfile(file_name):
        results = io.load_result_hdf5(file_name)
        const_3, slope_3, f_n_3, a_n_3, ph_n_3 = results['sin_mean']
        p_orb_3, _ = results['ephem']
        p_err_3, _ = results['ephem_err']
        model_linear = mdl.linear_curve(time, const_3, slope_3, i_chunks)
        model_sinusoid = mdl.sum_sines(time, f_n_3, a_n_3, ph_n_3)
        model_3 = model_linear + model_sinusoid
    else:
        const_3, slope_3, f_n_3, a_n_3, ph_n_3 = np.array([[], [], [], [], []])
        p_orb_3, p_err_3 = 0, 0
        model_3 = np.zeros(len(time))

    file_name = os.path.join(load_dir, f'{target_id}_analysis_4.hdf5')
    if os.path.isfile(file_name):
        results = io.load_result_hdf5(file_name)
        const_4, slope_4, f_n_4, a_n_4, ph_n_4 = results['sin_mean']
        model_linear = mdl.linear_curve(time, const_4, slope_4, i_chunks)
        model_sinusoid = mdl.sum_sines(time, f_n_4, a_n_4, ph_n_4)
        model_4 = model_linear + model_sinusoid
    else:
        const_4, slope_4, f_n_4, a_n_4, ph_n_4 = np.array([[], [], [], [], []])
        model_4 = np.zeros(len(time))

    file_name = os.path.join(load_dir, f'{target_id}_analysis_5.hdf5')
    if os.path.isfile(file_name):
        results = io.load_result_hdf5(file_name)
        const_5, slope_5, f_n_5, a_n_5, ph_n_5 = results['sin_mean']
        p_orb_5, _ = results['ephem']
        p_err_5, _ = results['ephem_err']
        t_tot, t_mean, t_mean_s, t_int, n_param_5, bic_5, noise_level_5 = results['stats']
        model_linear = mdl.linear_curve(time, const_5, slope_5, i_chunks)
        model_sinusoid = mdl.sum_sines(time, f_n_5, a_n_5, ph_n_5)
        model_5 = model_linear + model_sinusoid
        harmonics, harmonic_n = frs.find_harmonics_from_pattern(f_n_5, 1/p_orb_5, f_tol=1e-9)
        f_h_5, a_h_5, ph_h_5 = f_n_5[harmonics], a_n_5[harmonics], ph_n_5[harmonics]
    else:
        const_5, slope_5, f_n_5, a_n_5, ph_n_5 = np.array([[], [], [], [], []])
        p_orb_5, p_err_5 = 0, 0
        n_param_5, bic_5, noise_level_5 = 0, 0, 0
        model_5 = np.zeros(len(time))
        f_h_5, a_h_5, ph_h_5 = np.array([[], [], []])

    # stick together for sending to plot function
    models = [model_1, model_2, model_3, model_4, model_5]
    p_orb_i = [0, 0, p_orb_3, p_orb_3, p_orb_5]
    p_err_i = [0, 0, p_err_3, p_err_3, p_err_5]
    f_n_i = [f_n_1, f_n_2, f_n_3, f_n_4, f_n_5]
    a_n_i = [a_n_1, a_n_2, a_n_3, a_n_4, a_n_5]

    # plot frequency_analysis
    try:
        if save_dir is not None:
            file_name = os.path.join(save_dir, f'{target_id}_frequency_analysis_pd_full.png')
        else:
            file_name = None
        plot_pd_full_output(time, flux, flux_err, models, p_orb_i, p_err_i, f_n_i, a_n_i, i_chunks,
                                file_name=file_name, show=show)
        if np.any([len(fs) != 0 for fs in f_n_i]):
            plot_nr = np.arange(1, len(f_n_i) + 1)[[len(fs) != 0 for fs in f_n_i]][-1]
            plot_data = [eval(f'const_{plot_nr}'), eval(f'slope_{plot_nr}'),
                         eval(f'f_n_{plot_nr}'), eval(f'a_n_{plot_nr}'), eval(f'ph_n_{plot_nr}')]
            if save_dir is not None:
                file_name = os.path.join(save_dir, f'{target_id}_frequency_analysis_lc_sinusoids_{plot_nr}.png')
            else:
                file_name = None
            plot_lc_sinusoids(time, flux, *plot_data, i_chunks, file_name=file_name, show=show)
            if save_dir is not None:
                file_name = os.path.join(save_dir, f'{target_id}_frequency_analysis_pd_output_{plot_nr}.png')
            else:
                file_name = None
            plot_data = [p_orb_i[plot_nr - 1], p_err_i[plot_nr - 1]] + plot_data
            plot_pd_single_output(time, flux, flux_err, *plot_data, i_chunks, annotate=False,
                                      file_name=file_name, show=show)
            if save_dir is not None:
                file_name = os.path.join(save_dir, f'{target_id}_frequency_analysis_lc_harmonics_{plot_nr}.png')
            else:
                file_name = None
            plot_lc_harmonics(time, flux, *plot_data, i_chunks, file_name=file_name, show=show)
    except NameError:
        pass  # some variable wasn't loaded (file did not exist)
    except ValueError:
        pass  # no frequencies?

    return None

def plot_all_from_file(file_name, i_chunks=None, load_dir=None, save_dir=None, show=True):
    """Plot all diagnostic plots of the results for a given light curve file

    Parameters
    ----------
    file_name: str
        Path to a file containing the light curve data, with
        timestamps, normalised flux, error values as the
        first three columns, respectively.
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
    load_dir: str
        Path to a directory for loading analysis results.
        Will append <target_id> + _analysis automatically.
        Assumes the same directory as file_name if None.
    save_dir: str, None
        Path to a directory for saving the plots.
        Will append <target_id> + _analysis automatically.
        Directory is created if it doesn't exist yet.
    show: bool
        Whether to show the plots or not.

    Returns
    -------
    None
    """
    target_id = os.path.splitext(os.path.basename(file_name))[0]  # file name is used as target identifier
    if load_dir is None:
        load_dir = os.path.dirname(file_name)

    # load the data
    time, flux, flux_err = np.loadtxt(file_name, usecols=(0, 1, 2), unpack=True)

    # if sectors not given, take full length
    if i_chunks is None:
        i_chunks = np.array([[0, len(time)]])  # no sector information
    # i_half_s = i_chunks  # in this case no differentiation between half or full sectors

    # do the plotting
    sequential_plotting(time, flux, flux_err, i_chunks, target_id, load_dir, save_dir=save_dir, show=show)

    return None

def plot_all_from_tic(tic, all_files, load_dir=None, save_dir=None, show=True):
    """Plot all diagnostic plots of the results for a given light curve file

    Parameters
    ----------
    tic: int
        The TESS Input Catalog (TIC) number for loading/saving the data
        and later reference.
    all_files: list[str]
        List of all the TESS data product '.fits' files. The files
        with the corresponding TIC number are selected.
    load_dir: str
        Path to a directory for loading analysis results.
        Will append <tic> + _analysis automatically.
        Assumes the same directory as all_files if None.
    save_dir: str, None
        Path to a directory for saving the plots.
        Will append <tic> + _analysis automatically.
        Directory is created if it doesn't exist yet.
    show: bool
        Whether to show the plots or not.

    Returns
    -------
    None
    """
    if load_dir is None:
        load_dir = os.path.dirname(all_files[0])

    # load the data
    time, flux, flux_err, i_chunks, medians = io.load_light_curve(all_files, apply_flags=True)

    # do the plotting
    sequential_plotting(time, flux, flux_err, i_chunks, tic, load_dir, save_dir=save_dir, show=show)

    return None



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains functions for time series analysis;
specifically for the fitting of stellar oscillations and harmonic sinusoids.

Notes
-----
Minimize methods:
Nelder-Mead is extensively tested and found robust, while slow.
TNC is tested, and seems reliable while being fast, though slightly worse BIC results.
L-BFGS-B is tested, and seems reliable while being fast, though slightly worse BIC results.
See publication appendix for more information.

Code written by: Luc IJspeert
"""

import numpy as np
import numba as nb
from scipy.optimize import minimize

from star_shine.core import time_series as tms, model as mdl, goodness_of_fit as gof
from star_shine.core import frequency_sets as frs, periodogram as pdg


@nb.njit(cache=True)
def dsin_dx(two_pi_t, f, a, ph, d='f', f_base=0):
    """The derivative of a sine wave at times t, where x is on of the parameters.

    Parameters
    ----------
    two_pi_t: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series times two pi.
    f: float
        The frequency of a sine wave.
    a: float
        The amplitude of a sine wave.
    ph: float
        The phase of a sine wave.
    d: string
        Which derivative to take. Choose f, a, ph, h (for harmonics).
    f_base: float
        Base frequency of the harmonic series.

    Returns
    -------
    numpy.ndarray[Any, dtype[float]]
        Model time series of the derivative of a sine wave to f.

    Notes
    -----
    Make sure the phases correspond to the given time zero point.
    If d='h', it is assumed that f is a harmonic and the base frequency f_base is provided
    """
    if d == 'f':
        model_deriv = a * np.cos(two_pi_t * f + ph) * two_pi_t
    elif d == 'a':
        model_deriv = np.sin(two_pi_t * f + ph)
    elif d == 'ph':
        model_deriv = a * np.cos(two_pi_t * f + ph)
    elif d == 'h':
        model_deriv = a * np.cos(two_pi_t * f + ph) * two_pi_t * f / f_base  # (f / f_b = n_h)
    else:
        model_deriv = np.zeros(len(two_pi_t))

    return model_deriv


@nb.njit(cache=True)
def objective_sinusoids(params, time, flux, i_chunks, h_base, h_mult):
    """The objective function to give to scipy.optimize.minimize for a sum of sine waves.

    Parameters
    ----------
    params: numpy.ndarray[Any, dtype[float]]
        The parameters of a set of sine waves and linear curve(s)
        Has to be a flat array and are ordered in the following way:
        [f_base_1, f_base_2, ..., const_1, const_2, ..., slope_1, slope_2, ...,
        f_1, f_2, ..., a_1, a_2, ..., p_1, p_2, ...]
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
    h_base: numpy.ndarray[Any, dtype[int]]
         Indices of the base frequencies in params.
    h_mult: numpy.ndarray[Any, dtype[int]]
        Harmonic multiplier of the base frequency for each harmonic.

    Returns
    -------
    float
        Minus the (natural)log-likelihood of the residuals

    See Also
    --------
    linear_curve and sum_sines for the definition of the parameters.
    """
    n_chunk = len(i_chunks)  # each sector has its own slope (or two)
    n_base = len(h_base[h_mult == 1])
    n_sin = (len(params) - n_base - 2 * n_chunk) // 3  # each sine has freq, ampl and phase

    # separate the parameters
    f_base = params[:n_base]
    const = params[n_base:n_base + n_chunk]
    slope = params[n_base + n_chunk:n_base + 2 * n_chunk]
    f_n = params[n_base + 2 * n_chunk:n_base + 2 * n_chunk + n_sin]
    a_n = params[n_base + 2 * n_chunk + n_sin:n_base + 2 * n_chunk + 2 * n_sin]
    ph_n = params[n_base + 2 * n_chunk + 2 * n_sin:n_base + 2 * n_chunk + 3 * n_sin]

    # make the linear and sinusoid model
    model_linear = mdl.linear_curve(time, const, slope, i_chunks)
    model_sinusoid = mdl.sum_sines(time, f_n, a_n, ph_n)

    # interim residual
    resid = flux - model_linear - model_sinusoid

    # do the harmonics with linear regression (constant term should have already been taken care of)
    harmonic_betas, design_matrix, _ = solve_harmonic_sinusoids(time, resid, f_base[h_base], h_mult)

    # reconstruct the predicted flux based on our linear model and subtract
    resid -= design_matrix @ harmonic_betas

    # calculate the likelihood (minus this for minimisation)
    ln_likelihood = gof.calc_iid_normal_likelihood(resid)

    return -ln_likelihood


@nb.njit(cache=True, parallel=True)
def jacobian_sinusoids(params, time, flux, i_chunks, h_base, h_mult):
    """The jacobian function to give to scipy.optimize.minimize for a sum of sine waves.

    Parameters
    ----------
    params: numpy.ndarray[Any, dtype[float]]
        The parameters of a set of sine waves and linear curve(s)
        Has to be a flat array and are ordered in the following way:
        [f_base_1, f_base_2, ..., const_1, const_2, ..., slope_1, slope_2, ...,
        f_1, f_2, ..., a_1, a_2, ..., p_1, p_2, ...]
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
    h_base: numpy.ndarray[Any, dtype[int]]
         Indices of the base frequencies in params.
    h_mult: numpy.ndarray[Any, dtype[int]]
        Harmonic multiplier of the base frequency for each harmonic.

    Returns
    -------
    float
        The derivative of minus the (natural)log-likelihood of the residuals

    See Also
    --------
    objective_sinusoids
    """
    n_chunk = len(i_chunks)  # each sector has its own slope (or two)
    n_base = len(h_base[h_mult == 1])
    n_sin = (len(params) - n_base - 2 * n_chunk) // 3  # each sine has freq, ampl and phase

    # separate the parameters
    f_base = params[:n_base]
    const = params[n_base:n_base + n_chunk]
    slope = params[n_base + n_chunk:n_base + 2 * n_chunk]
    f_n = params[n_base + 2 * n_chunk:n_base + 2 * n_chunk + n_sin]
    a_n = params[n_base + 2 * n_chunk + n_sin:n_base + 2 * n_chunk + 2 * n_sin]
    ph_n = params[n_base + 2 * n_chunk + 2 * n_sin:n_base + 2 * n_chunk + 3 * n_sin]

    # make the linear and sinusoid model
    model_linear = mdl.linear_curve(time, const, slope, i_chunks)
    model_sinusoid = mdl.sum_sines(time, f_n, a_n, ph_n)

    # interim residual
    resid = flux - model_linear - model_sinusoid

    # do the harmonics with linear regression
    harmonic_betas, design_matrix, _ = solve_harmonic_sinusoids(time, resid, f_base[h_base], h_mult)

    # reconstruct the predicted flux based on our linear model and subtract
    resid -= design_matrix @ harmonic_betas

    two_pi_t = 2 * np.pi * (time - np.mean(time))

    # factor 1 of df/dx: -n / S
    df_1a = np.zeros(n_chunk)  # calculated per sector
    df_1b = -len(time) / np.sum(resid**2)

    # calculate the rest of the jacobian for the linear parameters, factor 2 of df/dx:
    df_2a = np.zeros(2 * n_chunk)
    for i in nb.prange(n_chunk):
        s = i_chunks[i]
        i_s = i + n_chunk
        df_1a[i] = -len(time[s[0]:s[1]]) / np.sum(resid[s[0]:s[1]]**2)
        df_2a[i] = np.sum(resid[s[0]:s[1]])
        df_2a[i_s] = np.sum(resid[s[0]:s[1]] * (time[s[0]:s[1]] - np.mean(time[s[0]:s[1]])))

    df_1a = np.append(df_1a, df_1a)  # copy to double length
    jac_lin = df_1a * df_2a

    # calculate the rest of the jacobian for the sinusoid parameters, factor 2 of df/dx:
    df_2b = np.zeros(3 * n_sin)
    for i in nb.prange(n_sin):
        i_f = n_base + i
        i_a = n_base + i + n_sin  # index of df_2b for a_n
        i_ph = n_base + i + 2 * n_sin  # index of df_2b for ph_n
        df_2b[i_f] = np.sum(resid * dsin_dx(two_pi_t, f_n[i], a_n[i], ph_n[i], d='f'))
        df_2b[i_a] = np.sum(resid * dsin_dx(two_pi_t, f_n[i], a_n[i], ph_n[i], d='a'))
        df_2b[i_ph] = np.sum(resid * dsin_dx(two_pi_t, f_n[i], a_n[i], ph_n[i], d='ph'))

    jac_sin = df_1b * df_2b

    # calculate the rest of the jacobian for the harmonic base frequencies, factor 2 of df/dx:
    f_h = f_base[h_base] * h_mult
    a_h = np.sqrt(harmonic_betas[1::2] ** 2 + harmonic_betas[2::2] ** 2)
    ph_h = np.arctan2(harmonic_betas[2::2], harmonic_betas[1::2])
    df_2c = np.zeros(n_base)
    for i in nb.prange(n_base):
        h_j = np.arange(len(h_base))[h_base == i]
        for j in h_j:
            df_2c[i] += np.sum(resid * dsin_dx(two_pi_t, f_h[j], a_h[j], ph_h[j], d='h', f_base=f_base[i]))

    jac_harm = df_1b * df_2c

    # jacobian = df/dx = df/dy * dy/dx (f is objective function, y is model)
    jac = np.concatenate((jac_harm, jac_lin, jac_sin))

    return jac


@nb.njit(cache=True)
def make_design_matrix_harmonic(time, f_base, h_mult, t_shift=True):
    """Generate a design matrix for a series of harmonics and a constant term.

    Also returns weights for each model feature.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series.
    f_base: numpy.ndarray[Any, dtype[float]]
        Base frequencies of the harmonics (of multiple series).
    h_mult: numpy.ndarray[Any, dtype[int]]
        Harmonic multiplier of the base frequency for each harmonic.
    t_shift: bool
        Mean center the time axis.

    Returns
    -------
    numpy.ndarray[Any, dtype[float]]
        The design matrix. N by M where N = len(time) and M = 1 + 2 len(h_mult).

    Notes
    -----
    Based on code from Puffins, by C. Johnston, D.W. Hogg, and N.L. Eisner;
    https://github.com/colej/puffins/
    Which is itself based on D. W. Hogg & S. Villar (2021, https://arxiv.org/pdf/2101.07256)
    """
    if t_shift:
        mean_t = np.mean(time)
    else:
        mean_t = 0

    # sinusoid angle term
    two_pi_f_t = 2. * np.pi * (f_base * h_mult)[None,:] * (time - mean_t)[:, None]

    # make empty matrix
    n_param = 1 + 2 * len(h_mult)
    design_matrix = np.ones((len(time), n_param))

    # constant term
    design_matrix[:, 0] = np.ones_like(time)
    # cosine terms
    design_matrix[:, 1::2] = np.cos(two_pi_f_t)
    # sine terms
    design_matrix[:, 2::2] = np.sin(two_pi_f_t)

    return design_matrix


@nb.njit(cache=True)
def make_feature_weights_harmonic(h_mult, s=0.1):
    """Generate the weights for the features of a harmonic series.

    Parameters
    ----------
    h_mult: numpy.ndarray[Any, dtype[int]]
        Harmonic multiplier of the base frequency for each harmonic (of multiple series).
    s: float
        The kernel width in Fourier space.

    Returns
    -------
    numpy.ndarray[Any, dtype[float]]
        Harmonic feature weights.

    Notes
    -----
    Based on code from Puffins, by C. Johnston, D.W. Hogg, and N.L. Eisner;
    https://github.com/colej/puffins/
    Which is itself based on D. W. Hogg & S. Villar (2021, https://arxiv.org/pdf/2101.07256)
    """
    # make empty array
    n_param = 1 + 2 * len(h_mult)
    feature_weights = np.zeros(n_param)

    # feature weights
    feature_weights[1::2] = (1. + s ** 2 * h_mult ** 2) ** 2
    feature_weights[2::2] = (1. + s ** 2 * h_mult ** 2) ** 2

    return feature_weights


@nb.njit(cache=True)
def solve_fw(x, y, w=None, l=None):
    """Feature weighted least-squares solver.

    Parameters
    ----------
    x: numpy.ndarray[(n, p), dtype[float]]
        Design matrix
    y: numpy.ndarray[(n,), dtype[float]]
        Target values
    l: numpy.ndarray[(p,), dtype[float]]
        Feature weights
    w: numpy.ndarray[(n,), dtype[float]]
        Weights for the targets
        Assumes matrix with diagonal elements as 1/inverse variance on Y

    Returns
    -------
    numpy.ndarray[(p,), dtype[float]]
        Solved coefficients (betas)

    Notes
    -----
    Implements equations 23 and 24 from D. W. Hogg & S. Villar (2021, https://arxiv.org/pdf/2101.07256)
    If weights is not provided, it assumes OLS
    W = 1/C^2, where C is the uncertainty on the target values

    Based on code from Puffins, by C. Johnston, D.W. Hogg, and N.L. Eisner;
    https://github.com/colej/puffins/
    """

    n, p = x.shape

    if w is None:
        w = np.ones(n)
    if l is None:
        l = np.ones(p)

    # Assumes the approximation of a periodic version of the Matern 3/2 kernal
    if n > p:  # Overdetermined
        # Use eq. 23 from https://arxiv.org/pdf/2101.07256
        # beta_hat = (X.T * W * X + L)^-1 * X.T * W * y

        xt_w = x.T * w
        xt_w_x = xt_w @ x
        # xt_w_x[np.diag_indices(p)] += l
        # Add l to the diagonal of xt_w_x
        np.fill_diagonal(xt_w_x, np.diag(xt_w_x) + l)

        # solve the linalg eq.
        betas = np.linalg.lstsq(xt_w_x, xt_w @ y, rcond=1e-14)[0]

    else:  # underdetermined
        # Use eq. 24 from https://arxiv.org/pdf/2101.07256
        # beta_hat = L^-1 @ X.T (X @ L^-1 @ X.T + 1/W)^-1 y

        linv_xt = (x / l).T
        x_linv_xt = x @ linv_xt
        # x_linv_xt[np.diag_indices(n)] += 1. / w
        # Add 1/w to the diagonal of x_linv_xt
        np.fill_diagonal(x_linv_xt, np.diag(x_linv_xt) + 1. / w)

        # solve the linalg eq.
        betas = linv_xt @ np.linalg.lstsq(x_linv_xt, y, rcond=1e-14)[0]

    return betas


@nb.njit(cache=True)
def solve_harmonic_sinusoids(time, flux, f_base, h_mult):
    """Solve for the amplitudes and phases of one or more harmonic series.

    Also returns the design matrix and feature weights.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series.
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series.
    f_base: numpy.ndarray[Any, dtype[float]]
        Base frequencies of the harmonics (of multiple series).
    h_mult: numpy.ndarray[Any, dtype[int]]
        Harmonic multiplier of the base frequency for each harmonic.

    Returns
    -------
    tuple
        numpy.ndarray[Any, dtype[float]]
            Harmonic amplitude solutions (cosine and sine amplitudes).
        numpy.ndarray[Any, dtype[float]]
            The design matrix. N by M where N = len(time) and M = 1 + 2 len(h_mult).
        numpy.ndarray[Any, dtype[float]]
            Harmonic feature weights.
    """
    # construct the design matrix associated with each datapoint-harmonic pair
    design_matrix = make_design_matrix_harmonic(time, f_base=f_base, h_mult=h_mult)

    # higher frequencies are penalised by weights
    feature_weights = make_feature_weights_harmonic(h_mult, s=0.01)  # use low s for weights that don't do much

    # solve the linalg for the model parameters
    harmonic_betas = solve_fw(design_matrix, flux, l=feature_weights, w=None)

    return harmonic_betas, design_matrix, feature_weights


def fit_multi_sinusoid_grouped(ts_model, g_min=45, g_max=50, logger=None):
    """Perform the multi-sinusoid, non-linear least-squares fit per sinusoid group.

    This will fit per groups of 45-50 sinusoids (by default), leaving the other sinusoid parameters fixed,
    reducing the overall runtime of the NL-LS fit.
    If present, harmonics are always fit simultaneously with each group.

    Parameters
    ----------
    ts_model: tms.TimeSeriesModel
        Instance of TimeSeriesModel containing the time series and model parameters.
    g_min: int
        Minimum group size
    g_max: int
        Maximum group size (g_max > g_min)
    logger: logging.Logger, optional
        Instance of the logging library.

    Notes
    -----
    ts_model must have been cleaned up before use here.
    """
    # get the harmonic base frequencies and harmonic parameters
    h_base_unique, h_base_map = ts_model.sinusoid.get_h_base_map()
    harmonics = ts_model.sinusoid.harmonics
    h_base = ts_model.sinusoid.h_base[harmonics]  # only include the harmonics
    h_mult = ts_model.sinusoid.h_mult[harmonics]  # only include the harmonics

    # indices of harmonic and non-harmonic sinusoids
    i_h = np.arange(len(harmonics))[harmonics]
    i_non_h = np.arange(len(harmonics))[~harmonics]

    # remove harmonics from the sinusoids (to be fitted)
    f_n = ts_model.sinusoid.f_n
    a_n = ts_model.sinusoid.a_n
    ph_n = ts_model.sinusoid.ph_n

    # prepare the group lists and some numbers
    f_groups = frs.group_frequencies_for_fit(a_n[i_non_h], g_min=g_min, g_max=g_max, indices=i_non_h)
    n_groups = len(f_groups)
    n_chunk = len(ts_model.i_chunks)
    n_sin = len(f_n)
    n_base = len(h_base_unique)

    # we don't want the frequencies to go lower than about 1/100/T
    f_low = 0.01 / ts_model.t_tot

    # prepare fixed fit input
    par_bounds = [(f_low, None) for _ in range(n_base)]  # harmonic series base frequencies
    par_bounds += [(None, None) for _ in range(2 * n_chunk)]  # constants and slopes

    # update the parameters for each group
    for k, group in enumerate(f_groups):
        # temporarily exclude group and harmonic sinusoids
        n_sin_g = len(group)
        ts_model.exclude_sinusoids(group)
        ts_model.exclude_sinusoids(i_h)

        # get current residual (excluding group, harmonics, and linear curve)
        residual = ts_model.flux - ts_model.sinusoid.sinusoid_model

        # get the periodogram amplitudes for amplitude limits
        pd_ampls, _ = pdg.scargle_ampl_phase(ts_model.time, residual, f_n[group])

        # prepare fit input
        par_init_i = np.concatenate((ts_model.sinusoid.f_base, ts_model.linear.const, ts_model.linear.slope,
                                     f_n[group], a_n[group], ph_n[group]))
        par_bounds_i = par_bounds + [(f_low, None) for _ in range(n_sin_g)]  # frequencies of free sinusoids
        par_bounds_i += [(0, 2 * pd_ampls[i]) for i in range(n_sin_g)]  # amplitudes of free sinusoids
        par_bounds_i += [(None, None) for _ in range(n_sin_g)]  # phases of free sinusoids
        arguments_i = (ts_model.time, residual, ts_model.i_chunks, h_base_map, h_mult)

        # do the fit
        result = minimize(objective_sinusoids, jac=jacobian_sinusoids, x0=par_init_i, args=arguments_i,
                          method='L-BFGS-B', bounds=par_bounds_i, options={'maxiter': 10 ** 4 * len(par_init_i)})

        # separate results
        f_base = result.x[:n_base]
        const = result.x[n_base:n_base + n_chunk]
        slope = result.x[n_base + n_chunk:n_base + 2 * n_chunk]
        f_g = result.x[n_base + 2 * n_chunk:n_base + 2 * n_chunk + n_sin_g]
        a_g = result.x[n_base + 2 * n_chunk + n_sin_g:n_base + 2 * n_chunk + 2 * n_sin_g]
        ph_g = result.x[n_base + 2 * n_chunk + 2 * n_sin_g:n_base + 2 * n_chunk + 3 * n_sin_g]

        # update model parameters (this re-includes the sinusoids)
        ts_model.set_linear_model(const, slope)
        ts_model.update_sinusoids(f_g, a_g, ph_g, group)

        # get the harmonic parameters with linear regression (using current residual, including group and linear curve)
        harmonic_betas, _, _ = solve_harmonic_sinusoids(ts_model.time, ts_model.residual(), f_base[h_base_map], h_mult)

        # get the harmonic amplitudes and phases from the betas
        cosines = harmonic_betas[1::2]
        sines = harmonic_betas[2::2]
        a_h = np.sqrt(cosines ** 2 + sines ** 2)
        ph_h = np.pi / 2 - np.arctan2(sines, cosines)  # this stupid formula ensures the phases are compatible

        # update the model parameters (this re-includes the harmonic sinusoids)
        f_h = f_base[h_base_map] * h_mult
        ts_model.update_sinusoids(f_h, a_h, ph_h, i_h, h_base_new=h_base, h_mult_new=h_mult)

    if logger is not None:
        logger.extra(f"N_f= {n_sin}, BIC= {ts_model.bic():1.2f} - Optimised model, N_group= {n_groups}",
                     extra={'update': True})

    return None



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains algorithms for data analysis.

Code written by: Luc IJspeert
"""

import numpy as np
import numba as nb
import itertools as itt

from star_shine.config import data_properties as dp


@nb.njit(cache=True)
def f_within_rayleigh(i, f_n, rayleigh):
    """Selects a chain of frequencies within the Rayleigh criterion from each other
    around the chosen frequency.

    Parameters
    ----------
    i: int
        Index of the frequency around which to search
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    rayleigh: float
        The appropriate frequency resolution (usually 1.5/T)

    Returns
    -------
    numpy.ndarray[Any, dtype[int]]
        Indices of close frequencies in the chain
    """
    indices = np.arange((len(f_n)))
    sorter = np.argsort(f_n)  # first sort by frequency
    f_diff = np.diff(f_n[sorter])  # spaces between frequencies
    sorted_pos = indices[sorter == i][0]  # position of i in the sorted array

    if np.all(f_diff > rayleigh):
        # none of the frequencies are close
        i_close = np.zeros(0, dtype=np.int_)
    elif np.all(f_diff < rayleigh):
        # all the frequencies are close
        i_close = indices
    else:
        # the frequency to investigate is somewhere inbetween other frequencies
        right_not_close = indices[sorted_pos + 1:][f_diff[sorted_pos:] > rayleigh]
        left_not_close = indices[:sorted_pos][f_diff[:sorted_pos] > rayleigh]

        # if any freqs to left or right are not close, take the first index where this happens
        if len(right_not_close) > 0:
            i_right_nc = right_not_close[0]
        else:
            i_right_nc = len(f_n)  # else take the right edge of the array
        if len(left_not_close) > 0:
            i_left_nc = left_not_close[-1]
        else:
            i_left_nc = -1  # else take the left edge of the array (minus one)

        # now select the frequencies close to f_n[i] by using the found boundaries
        i_close = indices[i_left_nc + 1:i_right_nc]

    # convert back to unsorted indices
    i_close_unsorted = sorter[i_close]

    return i_close_unsorted


@nb.njit(cache=True)
def chains_within_rayleigh(f_n, rayleigh):
    """Find all chains of frequencies within each other's Rayleigh criterion.

    Parameters
    ----------
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    rayleigh: float
        The appropriate frequency resolution (usually 1.5/T)

    Returns
    -------
    list[numpy.ndarray[Any, dtype[int]]]
        Indices of close frequencies in all found chains

    See Also
    --------
    f_within_rayleigh
    """
    indices = np.arange(len(f_n))
    used = []
    groups = []
    for i in indices:
        if i not in used:
            i_close = f_within_rayleigh(i, f_n, rayleigh)
            if len(i_close) > 1:
                used.extend(i_close)
                groups.append(i_close)

    return groups


@nb.njit(cache=True)
def f_close_pairs(f_n, df):
    """Find pairs of frequencies that are closer than df.

    Only retains the pair with the closest distance if a chain occurs.

    Parameters
    ----------
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    df: float
        The threshold difference below which two frequencies are considered too close.

    Returns
    -------
    list of lists
        A list where each sublist contains the indices of two frequencies from `f_n` that are closer than `df`.

    Examples
    --------
    In this example, the frequencies at indices 0 and 2 (i.e., 10.5 and 11.2), and at 1 and 3 are closer than 1.0.
    >>> f_n = np.array([10.5, 20.3, 11.2, 19.8])
    >>> df = 1.0
    >>> print(f_close_pairs(f_n, df))
    [[0, 2], [3, 1]]

    In this example, 0 and 2  are closer than 1.0, but 2 and 4 are closer than 0 and 2.
    >>> f_n = np.array([19.8, 10.1, 20.3, 11.2, 20.4])
    >>> df = 1.0
    >>> print(f_close_pairs(f_n, df))
    [[2, 4]]
    """
    # first sort by frequency
    sorter = np.argsort(f_n)

    # spaces between frequencies
    f_diff = np.diff(f_n[sorter])

    # find pairs that are too close
    pairs = []
    for i in range(len(f_n) - 1):
        if f_diff[i] < df:
            i_1, i_2 = sorter[i], sorter[i + 1]

            # loop through existing pairs to find double use
            remove_pair = []
            append = True
            for j in range(len(pairs)):
                if i_1 in pairs[j] or i_2 in pairs[j]:
                    if f_diff[i] < (f_n[pairs[j][1]] - f_n[pairs[j][0]]):
                        remove_pair.append(j)
                    else:
                        append = False

            for pair in remove_pair:
                pairs.pop(pair)

            if append:
                pairs.append([i_1, i_2])

    return pairs


@nb.njit(cache=True)
def remove_insignificant_sigma(f_n, f_n_err, a_n, a_n_err, sigma_f=1., sigma_a=3.):
    """Removes insufficiently significant frequencies in terms of error margins.

    Frequencies with an amplitude less than sigma times the error are removed, as well as those that have
    an overlapping frequency error and are lower amplitude than any of the overlapped frequencies.

    Parameters
    ----------
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    f_n_err: numpy.ndarray[Any, dtype[float]]
        Formal errors in the frequencies
    a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sine waves
    a_n_err: numpy.ndarray[Any, dtype[float]]
        Formal errors in the amplitudes
    sigma_a: float
        Number of times the error to use for check of significant amplitude
    sigma_f: float
        Number of times the error to use for check of significant
        frequency separation

    Returns
    -------
    numpy.ndarray[Any, dtype[bool]]
        Boolean mask of frequencies deemed insignificant
    """
    # amplitude not significant enough
    a_insig = (a_n / a_n_err < sigma_a)

    # frequency error overlaps with neighbour
    f_insig = np.zeros(len(f_n), dtype=np.bool_)
    for i in range(len(f_n)):
        overlap = (f_n[i] + sigma_f * f_n_err[i] > f_n) & (f_n[i] - sigma_f * f_n_err[i] < f_n)

        # if any of the overlap is higher in amplitude, throw this one out
        if np.any((a_n[overlap] > a_n[i]) & (f_n[overlap] != f_n[i])):
            f_insig[i] = True

    remove = a_insig | f_insig

    return remove


def remove_insignificant_snr(time, a_n, noise_at_f):
    """Removes insufficiently significant frequencies in terms of S/N.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sine waves
    noise_at_f: numpy.ndarray[Any, dtype[float]]
        The noise level at each frequency

    Returns
    -------
    numpy.ndarray[Any, dtype[int]]
        Indices of frequencies deemed insignificant

    Notes
    -----
    Frequencies with an amplitude less than the S/N threshold are removed,
    using a threshold appropriate for TESS as function of the number of
    data points.

    The noise_at_f here captures the amount of noise on fitting a
    sinusoid of a certain frequency to all data points.
    Not to be confused with the noise on the individual data points of the
    time series.
    """
    snr_threshold = dp.signal_to_noise_threshold(time)

    # signal-to-noise below threshold
    a_insig = (a_n / noise_at_f < snr_threshold)

    return a_insig


def group_frequencies_for_fit(a_n, g_min=20, g_max=25, indices=None):
    """Groups sinusoids into sets of g_min to g_max for multi-sine fitting

    Parameters
    ----------
    a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sine waves
    g_min: int
        Minimum group size
    g_max: int
        Maximum group size (g_max > g_min)
    indices: numpy.ndarray[Any, dtype[int]]
        Alternate sinusoid indices to use.

    Returns
    -------
    list[numpy.ndarray[Any, dtype[int]]]
        List of sets of indices indicating the groups

    Notes
    -----
    To make the task of fitting more manageable, the free parameters are binned into groups, in which
    the remaining parameters are kept fixed. Sinusoids of similar amplitude are grouped together,
    and the group cut-off is determined by the biggest gaps in amplitude between frequencies, but group size
    is always kept between g_min and g_max. g_min < g_max. The idea of using amplitudes is that sinusoids
    of similar amplitude have a similar amount of influence on each other.
    """
    # keep track of which freqs have been used with the sorted indices
    not_used = np.argsort(a_n)[::-1]

    groups = []
    while len(not_used) > 0:
        if len(not_used) > g_min + 1:
            # find index of maximum amplitude difference in group size range
            a_diff = np.diff(a_n[not_used[g_min:g_max + 1]])
            i_max = np.argmin(a_diff)  # the diffs are negative so this is max absolute difference

            # since not_used is sorted, pick the first i_group sinusoids
            i_group = g_min + i_max + 1
            group_i = not_used[:i_group]
        else:
            # we have minimum group size or fewer left, so use all
            group_i = np.copy(not_used)
            i_group = len(not_used)

        # delete group_i from not_used and append group_i to groups
        not_used = np.delete(not_used, np.arange(i_group))

        # if indices are provided, swap them in
        if indices is not None:
            groups.append(indices[group_i])
        else:
            groups.append(group_i)

    return groups


@nb.njit(cache=True)
def find_harmonics(f_n, f_n_err, f_base, sigma=1.):
    """Find the orbital harmonics from a set of frequencies, given the base frequency.

    Parameters
    ----------
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    f_n_err: numpy.ndarray[Any, dtype[float]]
        Formal errors in the frequencies
    f_base: float
        Base frequency of the harmonic series.
    sigma: float
        Number of times the error to use for check of significance

    Returns
    -------
    numpy.ndarray[bool]
        Indices of frequencies that are harmonics of p_orb

    Notes
    -----
    Only includes those frequencies that are within sigma * error of an orbital harmonic.
    If multiple frequencies correspond to one harmonic, only the closest is kept.
    """
    # the frequencies divided by the orbital frequency gives integers for harmonics
    test_int = f_n / f_base
    is_harmonic = ((test_int % 1) > 1 - sigma * f_n_err / f_base) | ((test_int % 1) < sigma * f_n_err / f_base)

    # look for harmonics that have multiple close frequencies
    harmonic_f = f_n[is_harmonic]
    sorter = np.argsort(harmonic_f)
    harmonic_n = np.round(test_int[is_harmonic], 0, np.zeros(np.sum(is_harmonic)))  # third arg needed for numba
    n_diff = np.diff(harmonic_n[sorter])

    # only keep the closest frequencies
    if np.any(n_diff == 0):
        n_dup = np.unique(harmonic_n[sorter][:-1][n_diff == 0])
        for n in n_dup:
            is_harmonic[np.round(test_int, 0, np.zeros(len(test_int))) == n] = False
            is_harmonic[np.argmin(np.abs(test_int - n))] = True

    i_harmonic = np.arange(len(f_n))[is_harmonic]

    return i_harmonic


@nb.njit(cache=True)
def construct_harmonic_range(f_0, domain):
    """create a range of harmonic frequencies given the base frequency.

    Parameters
    ----------
    f_0: float
        Base frequency in the range, from where the rest of the pattern is built.
    domain: list[float], numpy.ndarray[2, dtype[float]]
        Two values that give the borders of the range.
        Sensible values could be the Rayleigh criterion and the Nyquist frequency

    Returns
    -------
    tuple
        A tuple containing the following elements:
        harmonics: numpy.ndarray[Any, dtype[float]]
            Frequencies of the harmonic series in the domain
        n_range: numpy.ndarray[Any, dtype[int]]
            Corresponding harmonic numbers (base frequency is 1)
    """
    # determine where the range of harmonics starts and ends
    n_start = np.ceil(domain[0] / f_0)
    n_end = np.floor(domain[1] / f_0)
    n_range = np.arange(max(1, n_start), n_end + 1).astype(np.int_)
    harmonics = f_0 * n_range

    return harmonics, n_range


@nb.njit(cache=True)
def find_harmonics_from_pattern(f_n, f_base, f_tol=1e-9):
    """Get the indices of the frequencies matching closest to the harmonics.

    Parameters
    ----------
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    f_base: float
        Base frequency of the harmonic series.
    f_tol: float
        Tolerance in the frequency for accepting harmonics

    Returns
    -------
    tuple
        A tuple containing the following elements:
        harmonics: numpy.ndarray[Any, dtype[int]]
            Indices of the frequencies in f_n that are deemed harmonics
        harmonic_n: numpy.ndarray[Any, dtype[int]]
            Corresponding harmonic numbers (base frequency is 1)

    Notes
    -----
    A frequency is only accepted as harmonic if it is within 1e-9 of the pattern
    (by default). This can now be user defined for more flexibility.
    """
    # guard against zero period or empty list
    if (f_base == 0) | (len(f_n) == 0):
        harmonics = np.zeros(0, dtype=np.int_)
        harmonic_n = np.zeros(0, dtype=np.int_)
        return harmonics, harmonic_n

    # make the pattern of harmonics
    domain = [0, np.max(f_n) + 0.5 * f_base]
    harmonic_pattern, harmonic_n = construct_harmonic_range(f_base, domain)

    # sort the frequencies
    sorter = np.argsort(f_n)
    f_n = f_n[sorter]

    # get nearest neighbour in harmonics for each f_n by looking to the left and right of the sorted position
    i_nn = np.searchsorted(f_n, harmonic_pattern)
    i_nn[i_nn == len(f_n)] = len(f_n) - 1
    closest = np.abs(f_n[i_nn] - harmonic_pattern) < np.abs(harmonic_pattern - f_n[i_nn - 1])
    i_nn = i_nn * closest + (i_nn - 1) * np.invert(closest)

    # get the distances to nearest neighbours
    d_nn = np.abs(f_n[i_nn] - harmonic_pattern)

    # check that the closest neighbours are reasonably close to the harmonic
    m_cn = (d_nn < min(f_tol, f_base / 2))  # distance must be smaller than tolerance (and never larger than 1/2P)

    # keep the ones left over
    harmonics = sorter[i_nn[m_cn]]
    harmonic_n = harmonic_n[m_cn]

    return harmonics, harmonic_n


@nb.njit(cache=True)
def find_harmonics_tolerance(f_n, f_base, f_tol):
    """Get the indices of the frequencies matching within a tolerance to the harmonics.

    Parameters
    ----------
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    f_base: float
        Base frequency of the harmonic series.
    f_tol: float
        Tolerance in the frequency for accepting harmonics

    Returns
    -------
    tuple
        A tuple containing the following elements:
        harmonics: numpy.ndarray[Any, dtype[int]]
            Indices of the frequencies in f_n that are deemed harmonics
        harmonic_n: numpy.ndarray[Any, dtype[int]]
            Corresponding harmonic numbers (base frequency is 1)

    Notes
    -----
    A frequency is only accepted as harmonic if it is within some relative error.
    This can be user defined for flexibility.
    """
    harmonic_n = np.zeros(len(f_n))
    harmonic_n = np.round(f_n / f_base, 0, harmonic_n)  # closest harmonic (out argument needed in numba atm)
    harmonic_n[harmonic_n == 0] = 1  # avoid zeros resulting from large f_tol

    # get the distances to the nearest pattern frequency
    d_nn = np.abs(f_n - harmonic_n * f_base)

    # check that the closest neighbours are reasonably close to the harmonic
    m_cn = (d_nn < min(f_tol, f_base / 2))  # distance smaller than tolerance (or half the harmonic spacing)

    # produce indices and make the right selection
    harmonics = np.arange(len(f_n))[m_cn]
    harmonic_n = harmonic_n[m_cn].astype(np.int_)

    return harmonics, harmonic_n


@nb.njit(cache=True)
def select_harmonics_sigma(f_n, f_n_err, f_base, f_tol, sigma_f=3):
    """Selects only those frequencies that are probably harmonics

    Parameters
    ----------
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    f_n_err: numpy.ndarray[Any, dtype[float]]
        Formal errors in the frequencies
    f_base: float
        Base frequency of the harmonic series.
    f_tol: float
        Tolerance in the frequency for accepting harmonics
    sigma_f: float
        Number of times the error to use for check of significant
        frequency separation

    Returns
    -------
    tuple
        A tuple containing the following elements:
         harmonics_passed: numpy.ndarray[Any, dtype[int]]
            Indices of the frequencies in f_n that are harmonics
         harmonic_n: numpy.ndarray[Any, dtype[int]]
            Corresponding harmonic numbers (base frequency is 1)

    Notes
    -----
    A frequency is only accepted as harmonic if it is within the
    frequency resolution of the pattern, and if it is within <sigma_f> sigma
    of the frequency uncertainty
    """
    # get the harmonics within f_tol
    harmonics, harm_n = find_harmonics_from_pattern(f_n, f_base, f_tol=f_tol)

    # frequency error overlaps with theoretical harmonic
    passed_h = np.zeros(len(harmonics), dtype=np.bool_)
    for i, (h, n) in enumerate(zip(harmonics, harm_n)):
        f_theo = n * f_base
        margin = sigma_f * f_n_err[h]
        overlap_h = (f_n[h] + margin > f_theo) & (f_n[h] - margin < f_theo)
        if overlap_h:
            passed_h[i] = True

    harmonics_passed = harmonics[passed_h]
    harmonic_n = harm_n[passed_h]

    return harmonics_passed, harmonic_n


# @nb.njit()  # won't work due to itertools
def find_combinations(f_n, f_n_err, sigma=1.):
    """Find linear combinations from a set of frequencies.

    Parameters
    ----------
    f_n: list[float], numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    f_n_err: numpy.ndarray[Any, dtype[float]]
        Formal errors on the frequencies
    sigma: float
        Number of times the error to use for check of significance

    Returns
    -------
    tuple
        A tuple containing the following elements:
        final_o2: dict[int]
            Dictionary containing the indices of combinations of order 2
        final_o3: dict[int]
            Dictionary containing the indices of combinations of order 3

    Notes
    -----
    Only includes those frequencies that are within sigma * error of a linear combination.
    Does 2nd and 3rd order combinations. The number of sigma tolerance can be specified.
    """
    indices = np.arange(len(f_n))

    # make all combinations
    comb_order_2 = np.array(list(itt.combinations_with_replacement(indices, 2)))  # combinations of order 2
    comb_order_3 = np.array(list(itt.combinations_with_replacement(indices, 3)))  # combinations of order 2
    comb_freqs_o2 = np.sum(f_n[comb_order_2], axis=1)
    comb_freqs_o3 = np.sum(f_n[comb_order_3], axis=1)

    # check if any of the frequencies is a combination within error
    final_o2 = {}
    final_o3 = {}
    for i in indices:
        match_o2 = (f_n[i] > comb_freqs_o2 - sigma * f_n_err[i]) & (f_n[i] < comb_freqs_o2 + sigma * f_n_err[i])
        match_o3 = (f_n[i] > comb_freqs_o3 - sigma * f_n_err[i]) & (f_n[i] < comb_freqs_o3 + sigma * f_n_err[i])
        if np.any(match_o2):
            final_o2[i] = comb_order_2[match_o2]
        if np.any(match_o3):
            final_o3[i] = comb_order_3[match_o3]

    return final_o2, final_o3


def find_unknown_harmonics(f_n, f_n_err, sigma=1., n_max=5, f_tol=None):
    """Try to find harmonic series of unknown base frequency

    Parameters
    ----------
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    f_n_err: numpy.ndarray[Any, dtype[float]]
        Formal errors on the frequencies
    sigma: float
        Number of times the error to use for check of significance
    n_max: int
        Maximum divisor for each frequency in search of a base harmonic
    f_tol: None, float
        Tolerance in the frequency for accepting harmonics
        If None, use sigma matching instead of pattern matching

    Returns
    -------
    dict[int]
        Dictionary containing dictionaries with the indices of harmonic series

    Notes
    -----
    The first layer of the dictionary has the indices of frequencies as keys,
    the second layer uses n as keys and values are the indices of the harmonics
    stored in an array.
    n denotes the integer by which the frequency in question (index of the
    first layer) is divided to get the base frequency of the series.
    """
    indices = np.arange(len(f_n))
    n_harm = np.arange(1, n_max + 1)  # range of harmonic number that is tried for each frequency

    # test all frequencies for being the n-th harmonic in a series of harmonics
    candidate_h = {}
    for i in indices:
        for n in n_harm:
            f_base = f_n[i] / n
            if f_tol is not None:
                i_harmonic, _ = find_harmonics_from_pattern(f_n, f_base, f_tol=f_tol)
            else:
                i_harmonic = find_harmonics(f_n, f_n_err, f_base, sigma=sigma)  # harmonic indices

            if len(i_harmonic) > 1:
                # don't allow any gaps of more than 20 + the number of preceding harmonics
                set_i = np.arange(len(i_harmonic))
                set_sorter = np.argsort(f_n[i_harmonic])
                harm_n = np.rint(np.sort(f_n[i_harmonic][set_sorter]) / (f_n[i] / n))  # harmonic integer n
                large_gaps = (np.diff(harm_n) > 20 + set_i[:-1])
                if np.any(large_gaps):
                    cut_off = set_i[1:][large_gaps][0]
                    i_harmonic = i_harmonic[set_sorter][:cut_off]

                # only take sets that don't have frequencies lower than f_n[i]
                cond_1 = np.all(f_n[i_harmonic] >= f_n[i])
                cond_2 = (len(i_harmonic) > 1)  # check this again after cutting off gap
                if cond_1 & cond_2 & (i in candidate_h.keys()):
                    candidate_h[i][n] = i_harmonic
                elif cond_1 & cond_2:
                    candidate_h[i] = {n: i_harmonic}

    # check for conditions that only look at the set itself
    i_n_remove = []
    for i in candidate_h.keys():
        for n in candidate_h[i].keys():
            set_len = len(candidate_h[i][n])

            # determine the harmonic integer n of each frequency in the set
            harm_n = np.rint(np.sort(f_n[candidate_h[i][n]]) / (f_n[i] / n))

            # remove sets of two where n is larger than two
            cond_1 = (set_len == 2) & (n > 2)

            # remove sets of three where n is larger than three
            cond_2 = (set_len == 3) & (n > 3)

            # remove sets where all gaps are larger than three
            cond_3 = np.all(np.diff(harm_n) > 3)

            # remove sets with large gap between the first and second frequency
            cond_4 = (np.diff(harm_n)[0] > 7)

            # also remove any sets with n>1 that are not longer than the one with n=1
            if cond_1 | cond_2 | cond_3 | cond_4:
                if [i, n] not in i_n_remove:
                    i_n_remove.append([i, n])

    # remove entries
    for i, n in i_n_remove:
        candidate_h[i].pop(n, None)
        if len(candidate_h[i]) == 0:
            candidate_h.pop(i, None)

    # check whether a series is fully contained in another (and other criteria involving other sets)
    i_n_redundant = []
    for i in candidate_h.keys():
        for n in candidate_h[i].keys():
            # sets of keys to compare current (i, n) to
            compare = np.array([[j, k] for j in candidate_h.keys() for k in candidate_h[j].keys()
                                if (j != i) | (k != n)])

            # check whether this set is fully contained in another
            this_contained = np.array([np.all(np.in1d(candidate_h[i][n], candidate_h[j][k])) for j, k in compare])

            # check whether another set is fully contained in this one
            other_contained = np.array([np.all(np.in1d(candidate_h[j][k], candidate_h[i][n])) for j, k in compare])

            # check for equal length ones
            equal_length = np.array([len(candidate_h[i][n]) == len(candidate_h[j][k]) for j, k in compare])

            # those that are fully contained and same length are equal
            equal = (equal_length & this_contained)

            # remove equals from contained list
            this_contained = (this_contained & np.invert(equal))
            other_contained = (other_contained & np.invert(equal))

            # check for sets with the same starting f (or multiple)
            f_i, e_i = f_n[i] / n, f_n_err[i] / n  # the base frequency and corresponding error
            same_start = np.array([np.any((f_i + e_i * sigma > f_n[candidate_h[j][k]])
                                          & (f_i - e_i * sigma < f_n[candidate_h[j][k]])) for j, k in compare])

            # if this set is contained in another (larger) set and n is larger or equal, it is redundant
            for j, k in compare[this_contained]:
                if (k <= n) & ([i, n] not in i_n_redundant):
                    i_n_redundant.append([i, n])

            # if another set is contained in this (larger) one and n is larger, it is redundant
            for j, k in compare[other_contained]:
                if (k < n) & ([i, n] not in i_n_redundant):
                    i_n_redundant.append([i, n])

            # if this set is equal to another set but has higher n, it is redundant (we keep lowest n)
            for j, k in compare[equal]:
                if (k < n) & ([i, n] not in i_n_redundant):
                    i_n_redundant.append([i, n])

            # if this set starts with the same base frequency as another, but has higher n, it is redundant
            for j, k in compare[same_start]:
                if (k < n) & ([i, n] not in i_n_redundant):
                    i_n_redundant.append([i, n])

    # remove redundant entries
    for i, n in i_n_redundant:
        candidate_h[i].pop(n, None)
        if len(candidate_h[i]) == 0:
            candidate_h.pop(i, None)

    return candidate_h


@nb.njit(cache=True)
def harmonic_series_length(f_test, f_n, freq_res, f_max):
    """Find the number of harmonics that a set of frequencies has

    Parameters
    ----------
    f_test: numpy.ndarray[Any, dtype[float]]
        Frequencies to test at
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sine waves
    freq_res: float
        Frequency resolution
    f_max: float
        Highest allowed frequency at which signals are extracted

    Returns
    -------
    tuple
        A tuple containing the following elements:
        n_harm: numpy.ndarray[Any, dtype[float]]
            Number of harmonics per pattern
        completeness: numpy.ndarray[Any, dtype[float]]
            Completeness factor of each pattern
        distance: numpy.ndarray[Any, dtype[float]]
            Sum of squared distances between harmonics
    """
    n_harm = np.zeros(len(f_test))
    completeness = np.zeros(len(f_test))
    distance = np.zeros(len(f_test))

    for i, f in enumerate(f_test):
        harmonics, harmonic_n = find_harmonics_from_pattern(f_n, f, f_tol=freq_res / 2)
        n_harm[i] = len(harmonics)
        if n_harm[i] == 0:
            completeness[i] = 1
            distance[i] = 0
        else:
            completeness[i] = n_harm[i] / (f_max // f)
            distance[i] = np.sum((f_n[harmonics] - harmonic_n * f)**2)

    return n_harm, completeness, distance
